# coding: utf-8
"""Mambu loan Repayments objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

MambuRepayment holds a loan repayment. Don't instantiate this class
directly. Look at MambuRepayment pydoc for further info.

MambuRepayments holds a list of loan repayments.

Uses mambugeturl.getrepaymentsurl as default urlfunc
"""


from ..mambugeturl import getrepaymentsurl
from .mambustruct import MambuStruct
from .mamburestutils import MambuStructIterator

mod_urlfunc = getrepaymentsurl


class MambuRepayment(MambuStruct):
    """A loan Repayment from Mambu.

    Don't instantiate this class directly. It's mostly used by
    MambuRepayments to configure each of its elements as MambuRepayment
    objects. There's no suitable urlfunc to use to retrieve just a
    specific repayment from a loan account. In fact, you can look at the
    code of MambuRepayments.convert_dict_to_attrs(), it uses urlfunc and
    entid = None , so no connection to Mambu will be made, never, for
    any particular MambuRepayment object.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """Tasks done here:

        Just initializes the MambuStruct.
        """
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __repr__(self):
        """Instead of the default id given by the parent class, shows
        the duedate of the repayment.
        """
        return self.__class__.__name__ + " - duedate: %s" % self["dueDate"].strftime(
            "%Y-%m-%d"
        )


class MambuRepayments(MambuStruct):
    """A list of loan Repayments from Mambu.

    You SHOULD give an entid at instantiation time: the id for the loan
    account which repayments you wish to retrieve.

    Makes no sense to send an empty entid, which throws a
    MambuError.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """The trick for iterable Mambu Objects comes here:

        You iterate over each element of the responded List from Mambu,
        and create a Mambu Repayment object for each one, initializing
        them one at a time, and changing the attrs attribute (which just
        holds a list of plain dictionaries) with a MambuRepayment just
        created.

        You send a None urlfunc and entid to each MambuRepayment,
        because there's no method to retrieve individual repayments from
        Mambu, nor each individual repayment is identified by a
        particular ID.
        """
        for n, r in enumerate(self.attrs):
            # ok ok, I'm modifying elements of a list while iterating it. BAD PRACTICE!
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            try:
                self.mamburepaymentclass
            except AttributeError:
                self.mamburepaymentclass = MambuRepayment

            repayment = self.mamburepaymentclass(
                urlfunc=None, entid=None, *args, **kwargs
            )
            repayment.init(r, *args, **kwargs)
            repayment._MambuStruct__urlfunc = getrepaymentsurl
            self.attrs[n] = repayment



# coding: utf-8
"""Mambu Centres objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

MambuCentre holds a centre.

MambuCentres holds a list of centres.

Uses mambugeturl.getcentresurl as default urlfunc
"""


from ..mambugeturl import getcentresurl
from .mambustruct import MambuStruct
from .mamburestutils import MambuStructIterator

mod_urlfunc = getcentresurl


class MambuCentre(MambuStruct):
    """A Centre from Mambu.

    With the default urlfunc, entid argument must be the ID of the
    centre you wish to retrieve.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """Tasks done here:

        Just initializes the MambuStruct.
        """
        MambuStruct.__init__(
            self, urlfunc, entid, custom_field_name="customFieldValues", *args, **kwargs
        )

    def setBranch(self, *args, **kwargs):
        """Adds the branch for this centre to a 'branch' field.

                Returns the number of requests done to Mambu.
        """
        try:
            self.mambubranchclass
        except AttributeError:
            from .mambubranch import MambuBranch

            self.mambubranchclass = MambuBranch

        brnch = self.mambubranchclass(entid=self["assignedBranchKey"], *args, **kwargs)
        self["branch"] = brnch

        return 1


class MambuCentres(MambuStruct):
    """A list of Centres from Mambu.

    With the default urlfunc, entid argument must be empty at
    instantiation time to retrieve all the centres according to any
    other filter you send to the urlfunc.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """By default, entid argument is empty. That makes perfect
        sense: you want several centres, not just one
        """
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """The trick for iterable Mambu Objects comes here:

                You iterate over each element of the responded List from Mambu,
                and create a Mambu Centre object for each one, initializing them
                one at a time, and changing the attrs attribute (which just
                holds a list of plain dictionaries) with a MambuCentre just
                created.

        .. todo:: pass a valid (perhaps default) urlfunc, and its
                  corresponding id to entid to each MambuCentre, telling
                  MambuStruct not to connect() by default. It's desirable to
                  connect at any other further moment to refresh some element in
                  the list.
        """
        for n, b in enumerate(self.attrs):
            # ok ok, I'm modifying elements of a list while iterating it. BAD PRACTICE!
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            try:
                self.mambucentreclass
            except AttributeError:
                self.mambucentreclass = MambuCentre

            centre = self.mambucentreclass(urlfunc=None, entid=None, *args, **kwargs)
            centre.init(b, *args, **kwargs)
            centre._MambuStruct__urlfunc = getcentresurl
            self.attrs[n] = centre



# coding: utf-8
"""Mambu Branches objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

MambuBranch holds a branch.

MambuBranches holds a list of branches.

Uses mambugeturl.getbranchesurl as default urlfunc
"""


from ..mambugeturl import getbranchesurl
from .mambustruct import MambuStruct
from .mamburestutils import MambuStructIterator

mod_urlfunc = getbranchesurl


class MambuBranch(MambuStruct):
    """A Branch from Mambu.

    With the default urlfunc, entid argument must be the ID of the
    branch you wish to retrieve.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """Tasks done here:

        Just initializes the MambuStruct.
        """
        MambuStruct.__init__(
            self, urlfunc, entid, custom_field_name="customFieldValues", *args, **kwargs
        )

    def setUsers(self, *args, **kwargs):
        """Adds the active users for this branch to a 'users' field.

                Returns the number of requests done to Mambu.

        .. todo:: since pagination logic was added, is not always true that
                  just 1 request was done. It may be more! But since request
                  counter singleton holds true information about how many requests
                  were done to Mambu, in fact this return value may be obsolete
        """
        try:
            self.mambuuserclass
        except AttributeError:
            from .mambuuser import MambuUsers

            self.mambuusersclass = MambuUsers

        usrs = [
            us
            for us in self.mambuusersclass(branchId=self["id"], *args, **kwargs)
            if us["userState"] == "ACTIVE"
        ]
        self["users"] = usrs

        return 1

    def update(self, *args, **kwargs):
        """Updates a branch from Mambu
        """
        if "data" in kwargs:
            data = kwargs.pop("data")
            if data:
                raise NotImplementedError
        return super(MambuBranch, self).update(data={}, *args, **kwargs)


class MambuBranches(MambuStruct):
    """A list of Branches from Mambu.

    With the default urlfunc, entid argument must be empty at
    instantiation time to retrieve all the branches according to any
    other filter you send to the urlfunc.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """By default, entid argument is empty. That makes perfect
        sense: you want several branches, not just one
        """
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """The trick for iterable Mambu Objects comes here:

                You iterate over each element of the responded List from Mambu,
                and create a Mambu Branch object for each one, initializing them
                one at a time, and changing the attrs attribute (which just
                holds a list of plain dictionaries) with a MambuBranch just
                created.

        .. todo:: pass a valid (perhaps default) urlfunc, and its
                  corresponding id to entid to each MambuBranch, telling
                  MambuStruct not to connect() by default. It's desirable to
                  connect at any other further moment to refresh some element in
                  the list.
        """
        for n, b in enumerate(self.attrs):
            # ok ok, I'm modifying elements of a list while iterating it. BAD PRACTICE!
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            try:
                self.mambubranchclass
            except AttributeError:
                self.mambubranchclass = MambuBranch

            branch = self.mambubranchclass(urlfunc=None, entid=None, *args, **kwargs)
            branch.init(b, *args, **kwargs)
            branch._MambuStruct__urlfunc = getbranchesurl
            self.attrs[n] = branch



# coding: utf-8
"""Base class for all Mambu objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

Includes functionality to download such objects using GET requests to
Mambu, and to POST requests to Mambu.

Some caching may be achieved. Please look at caching done at
mambuproduct.AllMambuProducts class for an example of how caching should
be done.

Official Mambu dev documentation at Mambu Developer Center:
https://developer.mambu.com

You must configure your Mambu client in mambuconfig.py Please read the
pydocs there for more information.

Some basic definitions:

* MambuStruct refers to the parent class of all Mambu Objects and

* MambuPy Mambu objects refers to any child of MambuStruct, usually defined at
  some mambu-somename- python module file. Sometimes referred as implemented
  MambuStruct or something fancy

* Mambu entity refers to an entity retrieved via a request via Mambu REST
  API. It's a more abstract thing in fact. Also may refer to entities on a
  relational database but the term table is preferred in this case.
"""
import json
unicode = str
from datetime import datetime

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from copy import deepcopy

from ..mambuutil import (OUT_OF_BOUNDS_PAGINATION_LIMIT_VALUE, MambuCommError,
                         MambuError, apipwd, apiuser, encoded_dict, iri_to_uri,
                         strip_tags, setup_logging)
from .mamburestutils import RequestsCounter


logger = setup_logging(__name__)


class MambuStruct(object):
    """This one is the Father of all Mambu objects at MambuPy.

    It's a dictionary-like structure by default, self.attr attribute
    holds all the info respondend by Mambu REST API.

    This object is flexible enough to hold a list instead of a
    dictionary for certain objects holding iterators.

    Further work is needed to hold list-like behaviour however.
    """

    RETRIES = 5
    """This one holds the maximum number of retries for requests to Mambu.

    Some networks have a lot of lag, so a safe maximum has been added
    here. If after RETRIES attempts to connect to Mambu, MambuPy can't
    connect, a MambuCommError will be raised.
    """

    @staticmethod
    def serialize_fields(data):
        """Turns every attribute of the Mambu object in to a string representation.

        If the object is an iterable one, it goes down to each of its
        elements and turns its attributes too, recursively.

        The base case is when it's a MambuStruct class (this one) so it
        just 'serializes' the attr atribute. Refer to
        MambuStruct.serialize_struct pydoc.

        This is perhaps the worst way to do it, still looking for a better way.
        """
        if isinstance(data, MambuStruct):
            return data.serialize_struct()
        try:
            it = iter(data)
        except TypeError:
            return unicode(data)
        if type(it) == type(iter([])):
            l = []
            for e in it:
                l.append(MambuStruct.serialize_fields(e))
            return l
        elif type(it) == type(iter({})):
            d = {}
            for k in it:
                d[k] = MambuStruct.serialize_fields(data[k])
            return d
        # elif ... tuples? sets?
        return unicode(data)

    def __getitem__(self, key):
        """Dict-like key query"""
        return self.attrs[key]

    def __setitem__(self, key, value):
        """Dict-like set"""
        self.attrs[key] = value

    def __delitem__(self, key):
        """Dict-like del key"""
        del self.attrs[key]

    def __hash__(self):
        """Hash of the object"""
        try:
            return hash(self.id)
        except Exception:
            return hash(self.__repr__())

    def __getattribute__(self, name):
        """Object-like get attribute

        When accessing an attribute, tries to find it in the attrs
        dictionary, so now MambuStruct may act not only as a dict-like
        structure, but as a full object-like too (this is the getter
        side).
        """
        try:
            # first, try to read 'name' as if it's a property of the object
            # if it doesn't exists as property, AttributeError raises
            return object.__getattribute__(self, name)
        except AttributeError:
            # try to read the attrs property
            attrs = object.__getattribute__(self, "attrs")
            if type(attrs) == list or name not in attrs:
                # magic won't happen when not a dict-like MambuStruct or
                # when attrs has not the 'name' key (this last one means
                # that if 'name' is not a property of the object too,
                # AttributeError will raise by default)
                return object.__getattribute__(self, name)
            # all else, read the property from the attrs dict, but with a . syntax
            return attrs[name]

    def __setattr__(self, name, value):
        """Object-like set attribute

        When setting an attribute, tries to set it in the attrs
        dictionary, so now MambuStruct acts not only as a dict-like
        structure, but as a full object-like too (this is the setter
        side).
        """
        try:
            # attrs needs to exist to make the magic happen!
            # ... if not, AttributeError raises
            attrs = object.__getattribute__(self, "attrs")
            if type(attrs) == list:
                # when not treating with a dict-like MambuStruct...
                raise AttributeError
            try:
                # see if 'name' is currently a property of the object
                object.__getattribute__(self, name)
            except AttributeError:
                # if not, then assign it as a new key in the dict
                attrs[name] = value
            else:
                raise AttributeError

        # all else, assign it as a property of the object
        except AttributeError:
            object.__setattr__(self, name, value)

    def __repr__(self):
        """Mambu object repr tells the class name and the usual 'id' for it.

        If an iterable, it instead gives its length.
        """
        try:
            return self.__class__.__name__ + " - id: %s" % self.attrs["id"]
        except KeyError:
            return self.__class__.__name__ + " (no standard entity)"
        except AttributeError:
            return (
                self.__class__.__name__
                + " - id: '%s' (not synced with Mambu)" % self.entid
            )
        except TypeError:
            return self.__class__.__name__ + " - len: %s" % len(self.attrs)

    def __str__(self):
        """Mambu object str gives a string representation of the attrs attribute."""
        try:
            return self.__class__.__name__ + " - " + str(self.attrs)
        except AttributeError:
            return repr(self)

    def __len__(self):
        """Length of the attrs attribute.

        If dict-like (not iterable), it's the number of keys holded on the attrs dictionary.
        If list-like (iterable), it's the number of elements of the attrs list.
        """
        return len(self.attrs)

    def __eq__(self, other):
        """Very basic way to compare to Mambu objects.

                Only looking at their EncodedKey field (its primary key on the
                Mambu DB).

        .. todo:: a lot of improvements may be done here.
        """
        if isinstance(other, MambuStruct):
            try:
                if "encodedKey" not in other.attrs or "encodedKey" not in self.attrs:
                    return NotImplemented
            except AttributeError:
                return NotImplemented
            return other["encodedKey"] == self["encodedKey"]

    def has_key(self, key):
        """Dict-like behaviour"""
        try:
            if type(self.attrs) == dict:
                return key in self.attrs
            else:
                raise AttributeError  # if attrs is not a dict
        except AttributeError:  # if attrs doesnt exist
            raise NotImplementedError

    def __contains__(self, item):
        """Dict-like and List-like behaviour"""
        return item in self.attrs

    def get(self, key, default=None):
        """Dict-like behaviour"""
        if type(self.attrs) == dict:
            return self.attrs.get(key, default)
        else:
            raise NotImplementedError  # if attrs is not a dict

    def keys(self):
        """Dict-like behaviour"""
        try:
            return self.attrs.keys()
        except AttributeError:
            raise NotImplementedError

    def items(self):
        """Dict-like behaviour"""
        try:
            return self.attrs.items()
        except AttributeError:
            raise NotImplementedError

    def init(self, attrs={}, *args, **kwargs):
        """Default initialization from a dictionary responded by Mambu

        in to the elements of the Mambu object.

        It assings the response to attrs attribute and converts each of
        its elements from a string to an adequate python object: number,
        datetime, etc.

        Basically it stores the response on the attrs attribute, then
        runs some customizable preprocess method, then runs
        convert_dict_to_attrs method to convert the string elements to an
        adequate python object, then a customizable postprocess method.

        It also executes each method on the 'methods' attribute given on
        instantiation time, and sets new customizable 'properties' to
        the object.

        Why not on __init__? two reasons:

        * __init__ optionally connects to Mambu, if you don't connect to
          Mambu, the Mambu object will be configured but it won't have
          any Mambu info on it. Only when connected, the Mambu object
          will be initialized, here.

          Useful to POST several times the same Mambu object. You make a
          POST request over and over again by calling it's connect()
          method every time you wish. This init method will configure the
          response in to the attrs attribute each time.

          You may also wish to update the info on a previously initialized
          Mambu object and refresh it with what Mambu now has. Instead of
          building a new object, you just connect() again and it will be
          refreshed.

        * Iterable Mambu objects (lists) do not initialize here, the
          iterable Mambu object __init__ goes through each of its elements
          and then initializes with this code one by one. Please look at
          some Mambu iterable object code and pydoc for more details."""
        self.attrs = attrs
        self.preprocess()
        self.convert_dict_to_attrs(*args, **kwargs)
        self.postprocess()
        try:
            self.entid = self.id
        except (TypeError, AttributeError):
            pass
        try:
            for meth in kwargs["methods"]:
                try:
                    getattr(self, meth)()
                except Exception:
                    pass
        except Exception:
            pass
        try:
            for propname, propval in kwargs["properties"].items():
                setattr(self, propname, propval)
        except Exception:
            pass

    def serialize_struct(self):
        """Makes a string from each element on the attrs attribute.

                Read the class attribute MambuStruct.serialize_fields pydoc for
                more info.

                It DOES NOT serializes the class, for persistence or network
                transmission, just the fields on the attrs attribute.

                Remember that attrs may have any type of elements: numbers,
                strings, datetimes, Mambu objects, etc. This is a way to convert
                it to a string in an easy, however ugly, way.

                WARNING: it may fall in a stack overflow

        .. todo:: check recursion levels.
        """
        serial = MambuStruct.serialize_fields(self.attrs)
        return serial

    def __copy_args(self, args, kwargs):
        self.__args = deepcopy(args)
        for k, v in kwargs.items():
            self.__kwargs[k] = deepcopy(v)

    def __init__(self, urlfunc, entid="", *args, **kwargs):
        """Initializes a new Mambu object.

        Args:
         urlfunc (str): is the only required parameter. The urlfunc
                           returns a string of the URL to make a request to Mambu. You
                           may read about the valid urlfuncs supported by MambuPy at
                           mambuutil.py
         entid (str): is the usual ID of a Mambu entity you like to GET from
                         Mambu. The ID of a loan account, a client, a group, etc. Or the
                         transactions or repayments of certain loan account. It's an
                         optional parameter because the iterable Mambu objects don't need
                         an specific ID, just some other filtering parameters supported
                         on the urlfunc function.

        If you send a None urlfunc, the object will be configured, but
        won't connect to Mambu, never. Useful for iterables that
        configure their elements but can't or won't need Mambu to send
        further information. See :any:`mamburepayment.MambuRepayments`
        for an example of this.

        Should you need to add support for more functionality, add them
        at mambuutuil.

        Many other arguments may be sent here to further configure the
        Mambu object or the response by Mambu REST API:

        * **debug** flag (currently not implemented, just stored on the object)
        * **connect** flag, to optionally omit the connection to Mambu (see
          the init() (with no underscores) method pydoc)
        * **data** parameter for POST requests (see connect() method pydoc)
        * **date_format** parameter (see util_date_format() method pydoc)

        Also, parameters to be sent to Mambu on the request, such as:

        * **limit** parameter (for pagination, see connect() method
          pydoc). If 0 (default) tells connect to auto-paginate. All
          else is used as given for manual pagination (all along with
          the offset argument too)
        * **offset** parameter (for pagination, see connect() method
          pydoc)
        * **fullDetails**, accountState and other filtering parameters
          (see mambuutil pydocs)
        * **user**, **password**, **url** to connect to Mambu, to bypass
          mambuconfig configurations (see mambuutil pydoc)"""

        self.entid = entid
        """ID of the Mambu entity, or empty if not wanting a specific entity to be GETted"""

        self.rc = RequestsCounter()
        """Request Count Singleton"""

        connect = False
        """Default value connect"""

        try:
            self.__debug = kwargs.pop("debug")
            """Debug flag.

        .. todo:: not currently furtherly implemented
            """
        except KeyError:
            self.__debug = False

        try:
            self.__formato_fecha = kwargs.pop("date_format")
            """The default date format to be used for any datettime elements on the attrs attribute.

            Remember to use valid Python datetime strftime formats.
            """
        except KeyError:
            self.__formato_fecha = "%Y-%m-%dT%H:%M:%S+0000"

        try:
            self.__data = kwargs.pop("data")
            """JSON data to be sent to Mambu for POST/PATCH requests."""
        except KeyError:
            self.__data = None

        try:
            self.__method = kwargs.pop("method")
            """REST method to use when calling connect"""
        except KeyError:
            if self.__data:
                self.__method = "POST"
            else:
                self.__method = "GET"

        try:
            self.__limit = kwargs["limit"]
            self.__inilimit = self.__limit
            """Limit the number of elements to be retrieved from Mambu on GET requests.

            Defaults to 0 which means (for the connect method) that you
            don't care how many elements Mambu has, you wish to retrieve
            them all.
            """
            kwargs.pop("limit")
        except KeyError:
            self.__limit = 0

        try:
            self.__offset = kwargs["offset"]
            """When retrieving several elements from Mambu on GET
            requests, offset for the first element to be retrieved.
            """
            kwargs.pop("offset")
        except KeyError:
            self.__offset = 0

        try:
            self.custom_field_name = kwargs.pop("custom_field_name")
            """custom_field_name attribute."""
        except KeyError:
            pass

        # Basic authentication
        try:
            self.__user = kwargs.pop("user")
        except (KeyError, AttributeError):
            self.__user = apiuser
        try:
            self.__pwd = kwargs.pop("pwd")
        except (KeyError, AttributeError):
            self.__pwd = apipwd

        self.__headers = {"Accept": "application/vnd.mambu.v1+json"}

        try:
            if kwargs.pop("connect"):
                connect = True
        except KeyError:
            connect = True

        self.__args = ()
        self.__kwargs = {}

        self.__copy_args(args, kwargs)

        self.__urlfunc = urlfunc
        """The given urlfunc argument is saved here.

        It's used at the connect() method, when called.
        """
        if (
            self.__urlfunc == None
        ):  # Only used when GET returns an array, meaning the MambuStruct must be a list of MambuStucts
            return  # and each element is init without further configs. EDIT 2015-07-11: Really?

        if connect:
            self.connect(*args, **kwargs)

    def __request(self, url, user, pwd):
        """Method used in connect, for make the request.

        Uses self.__data dictionary to feed the body for POST and PATCH requests.
        If there is no data in self.__data, it defaults to a GET request.

        Args:
            url (str): url to make request, changed with iri_to_uri
            user (str): user to authenticate for requests
            pwd (str): password to authenticate for requests

        Returns:
            The response of the request (json)
        """
        retry_strategy = Retry(
            total=MambuStruct.RETRIES,
            status_forcelist=[429, 502, 503, 504],
            backoff_factor=1,
            allowed_methods=[
                "PUT",
                "DELETE",
                "POST",
                "PATCH"]
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        http = requests.Session()
        http.mount("https://", adapter)
        http.mount("http://", adapter)

        logger.debug(
            "%s %s, about to make %s request: url %s, data %s, headers %s",
            self.__class__.__name__, self.entid,
            self.__method, url, self.__data,
            [(k, v) for k, v in self.__headers.items()])
        if self.__data:
            data = json.dumps(encoded_dict(self.__data))
            self.__headers["Content-Type"] = "application/json"
            # PATCH
            if self.__method == "PATCH":
                resp = http.patch(
                    url, data=data, headers=self.__headers, auth=(user, pwd)
                )
            # POST
            else:
                resp = http.post(
                    url, data=data, headers=self.__headers, auth=(user, pwd)
                )
        else:
            # DELETE
            if self.__method == "DELETE":
                resp = http.delete(
                    url, headers=self.__headers, auth=(user, pwd)
                )
            # GET
            else:
                resp = http.get(
                    url, headers=self.__headers, auth=(user, pwd)
                )
        # Always count a new request when done!
        self.rc.add(datetime.now())

        logger.debug(
            "%s %s,response code to %s: %s",
            self.__class__.__name__, self.entid,
            self.__method, resp
        )
        logger.debug(
            "%s %s,response to %s: %s",
            self.__class__.__name__, self.entid,
            self.__method, resp.content
        )

        return resp

    def __process_jsonresp(self, jsonresp, limit, jsresp):
        window = True
        # Returns list: extend list for offset
        if type(jsonresp) == list:
            try:
                jsresp.extend(jsonresp)
            except AttributeError:
                # First window, forget that jsresp was a dict, turn it in to a list
                jsresp = jsonresp
            if len(jsonresp) < limit:
                window = False
        # Returns dict: in theory Mambu REST API doesn't takes limit/offset in to account
        else:
            jsresp = jsonresp
            self._is_mambu_error(jsresp)
            window = False

        return jsresp, window

    def __request_and_process(self, jsresp, url, user, pwd, limit, offset):
        resp = ""
        try:
            # Return a response, default Get
            resp = self.__request(url, user, pwd)
            resp.raise_for_status()

            self._raw_response_data = resp.content
            jsonresp = json.loads(resp.content)
            jsresp, window = self.__process_jsonresp(jsonresp, limit, jsresp)
        except requests.exceptions.HTTPError as httpex:
            logger.error(
                "%s %s, HTTPError on %s request: url %s, data %s, headers %s",
                self.__class__.__name__, self.entid,
                self.__method, url, self.__data,
                [(k, v) for k, v in self.__headers.items()])
            logger.error(
                "%s %s, %s: HTTPError: %s",
                self.__class__.__name__, self.entid, resp, str(httpex))
            if hasattr(resp, "content"):
                logger.error(
                    "%s %s, HTTPError, resp content: %s",
                    self.__class__.__name__, self.entid, resp.content)
            try:
                jsresp = json.loads(resp.content)
            except ValueError:
                # json.loads invalid data argument
                cont = resp.content
                raise MambuError(
                    strip_tags(cont.decode())
                    .strip()
                    .replace("\n\n", ": ", 1)
                    .replace("\n", ". ", 1)
                    .replace("\n", " ")
                    )
            except Exception as ex:
                # any other json error
                raise MambuError("JSON Error: %s" % repr(ex))

            try:
                self._is_mambu_error(jsresp)
            except AttributeError:  # pragma: no cover
                pass
        except requests.exceptions.RetryError as rerr:  # pragma: no cover
            logger.error(
                "%s %s, MambuCommError on %s request: url %s, data %s, headers %s",
                self.__class__.__name__, self.entid,
                self.__method, url, self.__data,
                [(k, v) for k, v in self.__headers.items()])
            logger.error(
                "%s %s, MambuCommError: %s",
                self.__class__.__name__, self.entid, str(rerr))
            raise MambuCommError(
                "ERROR I can't communicate with Mambu: {}".format(str(rerr)))
        except Exception as ex:
            """unknown exception"""
            logger.error(
                "%s %s, Exception on %s request: url %s, data %s, headers %s",
                self.__class__.__name__, self.entid,
                self.__method, url, self.__data,
                [(k, v) for k, v in self.__headers.items()])
            logger.exception(
                "%s %s, %s: Exception: %s",
                self.__class__.__name__, self.entid, resp, str(ex))
            if hasattr(resp, "content"):
                logger.error(
                    "%s %s, Exception, resp content: %s",
                    self.__class__.__name__, self.entid, resp.content)
            raise ex

        return jsresp, window

    def _is_mambu_error(self, jsresp):
        if (
            "returnCode" in jsresp and
            "returnStatus" in jsresp and
            jsresp["returnCode"] != 0
        ):
            raise MambuError(jsresp["returnStatus"])
        return False

    def __credentials(self):
        try:
            user = self.__user
        except AttributeError:
            self.__user = apiuser
            user = self.__user
        try:
            pwd = self.__pwd
        except AttributeError:
            self.__pwd = apipwd
            pwd = self.__pwd

        return user, pwd

    def connect(self, *args, **kwargs):
        """Connect to Mambu, make the request to the REST API.

                If there's no urlfunc to use, nothing is done here.

                When done, initializes the attrs attribute of the Mambu object
                by calling the init method. Please refer to that code and pydoc
                for further information.

                Uses urllib module to connect. Since all Mambu REST API
                responses are json, uses json module to translate the response
                to a valid python object (dictionary or list).

                When Mambu returns a response with returnCode and returnStatus
                fields, it means something went wrong with the request, and a
                MambuError is thrown detailing the error given by Mambu itself.

                If you need to make a POST request, send a data argument to the
                new Mambu object.

                Provides to prevent errors due to using special chars on the
                request URL. See mambuutil.iri_to_uri() method pydoc for further
                info.

                Provides to prevent errors due to using special chars on the
                parameters of a POST request. See mambuutil.encoded_dict()
                method pydoc for further info.

                For every request done, the request counter Singleton is
                increased.

                Includes retry logic, to provide for a max number of connection
                failures with Mambu. If maximum retries are reached,
                MambuCommError is thrown.

                Includes pagination code. Mambu supports a max of 500 elements
                per response. Such an ugly detail is wrapped here so further
                pagination logic is not needed above here. You need a million
                elements? you got them by making several 500 elements requests
                later joined together in a sigle list. Just send a limit=0
                (default) and that's it.

        .. todo:: improve raised exception messages. Sometimes
                  MambuCommErrors are thrown due to reasons not always
                  clear when catched down the road, but that perhaps may
                  be noticed in here and aren't fully reported to the
                  user. Specially serious on retries-MambuCommError
                  situations (the except Exception that increases the
                  retries counter is not saving the exception message,
                  just retrying).

        .. todo:: what about using decorators for the retry and for the
                  window logic?
                  (https://www.oreilly.com/ideas/5-reasons-you-need-to-learn-to-write-python-decorators
                  # Reusing impossible-to-reuse code)
        """
        self.__copy_args(args, kwargs)

        if not self.__urlfunc:
            return

        # Pagination window, Mambu restricts at most 500 elements in response
        offset = self.__offset
        window = True
        jsresp = {}
        while window:
            if not self.__limit or self.__limit > OUT_OF_BOUNDS_PAGINATION_LIMIT_VALUE:
                limit = OUT_OF_BOUNDS_PAGINATION_LIMIT_VALUE
            else:
                limit = self.__limit

            user, pwd = self.__credentials()

            urlfunc_params = self.__kwargs.copy()
            if self.__method == "GET" and not self.__data:
                urlfunc_params["limit"] = limit
                urlfunc_params["offset"] = offset
            url = iri_to_uri(
                self.__urlfunc(
                    str(self.entid),
                    *self.__args,
                    **urlfunc_params
                )
            )
            self.__url = url

            jsresp, window = self.__request_and_process(
                jsresp, url, user, pwd, limit, offset)

            # next window, moving offset...
            offset = offset + limit
            if self.__limit:
                self.__limit -= limit
                if self.__limit <= 0:
                    window = False
                    self.__limit = self.__inilimit

        if ("documents" not in url) and self.__method not in ["PATCH", "DELETE"]:
            if type(iter(jsresp)) == type(iter([])):
                logger.debug(
                    "%s, url %s, %s retrieved",
                    self.__class__.__name__, url, len(jsresp)
                )
            self.init(attrs=jsresp, *self.__args, **self.__kwargs)

    def _request_with_method(self, method, data, *args, **kwargs):
        """Method to do a specific request on the current Mambu entity"""
        self._MambuStruct__method = method
        self._MambuStruct__data = data
        self.connect(*args, **kwargs)
        self._MambuStruct__method = "GET"
        self._MambuStruct__data = None
        return 1

    def __massage_appropiate_format(self):
        if self.has_key(self.custom_field_name):
            self[self.custom_field_name] = [
                c
                for c in self[self.custom_field_name]
                if c["customField"]["state"] != "DEACTIVATED"
            ]
            for custom in self[self.custom_field_name]:
                field_name = custom["customField"]["name"]
                field_id = custom["customField"]["id"]
                if custom["customFieldSetGroupIndex"] != -1:
                    field_name += "_" + str(custom["customFieldSetGroupIndex"])
                    field_id += "_" + str(custom["customFieldSetGroupIndex"])
                custom["name"] = field_name
                custom["id"] = field_id
                try:
                    self[field_name] = custom["value"]
                    self[field_id] = custom["value"]
                except KeyError:
                    self[field_name] = custom["linkedEntityKeyValue"]
                    self[field_id] = custom["linkedEntityKeyValue"]
                    custom["value"] = custom["linkedEntityKeyValue"]

    def _process_fields(self):
        """Default info massage to appropiate format/style.

        This processing is called on preprocess and postprocess, AKA
        before and after conversion of fields to appropiate
        format/style.

        Perfect example: custom fields on certain objects is a mess
        (IMHO) when retrieved from Mambu, so some easiness is
        implemented here to access them. See some of this objects
        modules and pydocs for further info.

        Tasks done here:

        - Each custom field is given a 'name' key that holds the field
          name, and for each keyed name, the value of the custom field is
          assigned. Each pair of custom field name/value is entered as a
          new property on the main dictionary, allowing an easy access to
          them, not nested inside a pretty dark 'customInformation/Values'
          list.

        - Every item on the attrs dictionary gets stripped from trailing
          spaces (useful when users make typos).

        PLEASE REMEMBER! whenever you call postprocess on inherited
        classes you should call this method too, or else you lose the
        effect of the tasks done here.
        """
        try:
            try:
                self.__massage_appropiate_format()
            # in case you don't have any custom_field_name, don't do anything here
            except (AttributeError, TypeError):
                pass

            for k, v in self.items():
                try:
                    self[k] = v.strip()
                except Exception:
                    pass
        except NotImplementedError:
            pass

    def preprocess(self):
        """Each MambuStruct implementation may massage the info on the
        Mambu response before conversion to an appropiate format/style
        adequate for its needs.
        """
        self._process_fields()

    def postprocess(self):
        """Each MambuStruct implementation may massage the info on the
        Mambu response after conversion to an appropiate format/style
        adequate for its needs.
        """
        self._process_fields()

    def _convert_data_to_pytype(self, data):
        """"Python built-in types: ints, floats, or even datetimes.
        If it cannot convert it to a built-in type, leave it as string,
        or as-is. There may be nested Mambu objects here!
        This are the recursion base cases!
        """
        try:
            d = int(data)
            if (
                str(d) != data
            ):  # if string has trailing 0's, leave it as string, to not lose them
                return data
            return d
        except (TypeError, ValueError):
            try:
                return float(data)
            except (TypeError, ValueError):
                try:
                    return self.util_date_format(data)
                except (TypeError, ValueError):
                    return data
        return data

    def _convert_dict_to_pytypes(self, data):
        """Recursively convert the fields on the data given to a python object."""

        constantFields = [
            "id",
            "groupName",
            "name",
            "homePhone",
            "mobilePhone1",
            "phoneNumber",
            "postcode",
            "emailAddress",
            "description",
        ]
        # Iterators, lists and dictionaries
        # Here comes the recursive calls!
        try:
            it = iter(data)
            if type(it) == type(iter({})):
                d = {}
                for k in it:
                    if k in constantFields:
                        d[k] = data[k]
                    else:
                        d[k] = self._convert_dict_to_pytypes(data[k])
                data = d
            if type(it) == type(iter([])):
                l = []
                for e in it:
                    l.append(self._convert_dict_to_pytypes(e))
                data = l
        except TypeError:
            pass
        except Exception as ex:  # pragma: no cover
            """unknown exception"""
            raise ex
        return self._convert_data_to_pytype(data)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """Each element on the attrs attribute gest converted to a
        proper python object, depending on type.

        Some default constantFields are left as is (strings), because
        they are better treated as strings.
        """
        self.attrs = self._convert_dict_to_pytypes(self.attrs)

    def util_date_format(self, field, formato=None):
        """Converts a datetime field to a datetime using some specified format.

        What this really means is that, if specified format includes
        only for instance just year and month, day and further info gets
        ignored and the objects get a datetime with year and month, and
        day 1, hour 0, minute 0, etc.

        A useful format may be %Y%m%d, then the datetime objects
        effectively translates into date objects alone, with no relevant
        time information.

        PLEASE BE AWARE, that this may lose useful information for your
        datetimes from Mambu. Read this for why this may be a BAD idea:
        https://julien.danjou.info/blog/2015/python-and-timezones
        """
        if not formato:
            try:
                formato = self.__formato_fecha
            except AttributeError:
                formato = "%Y-%m-%dT%H:%M:%S+0000"
        return datetime.strptime(
            datetime.strptime(field, "%Y-%m-%dT%H:%M:%S+0000").strftime(formato),
            formato,
        )

    def create(self, data, *args, **kwargs):
        """Creates an entity in Mambu

        This method must be implemented in child classes

        Args:
            data (dictionary): dictionary with data to send, this dictionary
                               is specific for each Mambu entity
        """
        # if module of the function is diferent from the module of the object
        # that means create is not implemented in child class
        if self.create.__func__.__module__ != self.__module__:
            raise Exception("Child method not implemented")

        return self._request_with_method("POST", data, *args, **kwargs)

    def update(self, data, *args, **kwargs):
        """Downloads an update made to an entity in Mambu.

        This method must be implemented in child classes.

        Args:
            data (dictionary): data to send
        """
        if self.update.__func__.__module__ != self.__module__:
            raise NotImplementedError

        self.connect(*args, **kwargs)
        return 1

    def update_patch(self, data, *args, **kwargs):
        """Updates an entity in Mambu

        This method must be implemented in child classes

        Args:
            data (dictionary): dictionary with data to send, this dictionary
                               is specific for each Mambu entity
        """
        # if module of the function is diferent from the module of the object
        # that means create is not implemented in child class
        if self.update_patch.__func__.__module__ != self.__module__:
            raise NotImplementedError

        if not data:
            raise Exception("Requires data to update")

        return self._request_with_method("PATCH", data, *args, **kwargs)

    def update_post(self, data, *args, **kwargs):
        """Updates an entity in Mambu

        This method must be implemented in child classes

        Args:
            data (dictionary): dictionary with data to send, this dictionary
                               is specific for each Mambu entity
        """
        # if module of the function is diferent from the module of the object
        # that means create is not implemented in child class
        if self.update_post.__func__.__module__ != self.__module__:
            raise Exception("Child method not implemented")
        if not data:
            raise Exception("Requires data to update")

        return self._request_with_method("POST", data, *args, **kwargs)

    def upload_document(self, data, *args, **kwargs):
        """Uploads a document in Mambu

        This method must be implemented in child classes
        https://support.mambu.com/docs/attachments-api#post-attachments

        Args:
            data (dictionary): dictionary with data to send
        Example:
        data = {
            "document":{
                "documentHolderKey"     : self.encodedKey,
                "documentHolderType"    : "LOAN_ACCOUNT", # CLIENT, GROUP, USER, BRANCH...
                "name"                  : "loan_resume",
                "type"                  : "pdf",
            },
            "documentContent"           : "['encodedBase64_file']",
        }
        """
        # if module of the function is diferent from the module of the object
        # that means create is not implemented in child class
        if self.upload_document.__func__.__module__ != self.__module__:
            raise Exception("Child method not implemented")
        if not data:
            raise Exception("Requires data to upload")

        return self._request_with_method("POST", data, *args, **kwargs)



# coding: utf-8
"""Mambu loan Transaction objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

MambuTransaction holds a loan transaction. Don't instantiate this class
directly. Look at MambuTransaction pydoc for further info.

MambuTransactions holds a list of loan transactions.

Uses mambugeturl.gettransactionssurl as default urlfunc
"""


from ..mambugeturl import gettransactionsurl
from .mambustruct import MambuStruct
from .mamburestutils import MambuStructIterator

mod_urlfunc = gettransactionsurl


class MambuTransaction(MambuStruct):
    """A loan Transaction from Mambu.

    Don't instantiate this class directly. It's mostly used by
    MambuTransactions to configure each of its elements as
    MambuTransaction objects. There's no suitable urlfunc to use to
    retrieve just a specific transaction from a loan account. In fact,
    you can look at the code of MambuTransactions.convert_dict_to_attrs(),
    it uses urlfunc and entid = None , so no connection to Mambu will be
    made, never, for any particular MambuTransaction object.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """Tasks done here:

        Just initializes the MambuStruct.
        """
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __repr__(self):
        """Instead of the default id given by the parent class, shows
        the transactionid of the transaction.
        """
        try:
            return (
                self.__class__.__name__ + " - transactionid: %s" % self["transactionid"]
            )
        except KeyError:
            return (
                self.__class__.__name__ + " - transactionid: %s" % self["transactionId"]
            )


class MambuTransactions(MambuStruct):
    """A list of loan Transactions from Mambu.

    You SHOULD give an entid at instantiation time: the id for the loan
    account which transactions you wish to retrieve.

    Makes no sense to send an empty entid, which throws a
    MambuError.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """The trick for iterable Mambu Objects comes here:

                You iterate over each element of the responded List from Mambu,
                and create a Mambu Transaction object for each one, initializing
                them one at a time, and changing the attrs attribute (which just
                holds a list of plain dictionaries) with a MambuTransaction just
                created.

                You send a None urlfunc and entid to each MambuTransaction,
                because there's no method to retrieve individual transactions
                from Mambu.

        .. todo:: use the transactionid as a possible entid here.
        """
        for n, t in enumerate(self.attrs):
            # ok ok, I'm modifying elements of a list while iterating it. BAD PRACTICE!
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            try:
                self.mambutransactionclass
            except AttributeError:
                self.mambutransactionclass = MambuTransaction

            trans = self.mambutransactionclass(urlfunc=None, entid=None, *args, **kwargs)
            trans.init(t, *args, **kwargs)
            trans._MambuStruct__urlfunc = gettransactionsurl
            self.attrs[n] = trans



"""MambuPy's REST package (SOON TO BE DEPRECATED).

Objects using `Mambu REST API
<https://developer.mambu.com/customer/en/portal/articles/1162276-rest-apis-overview/>`_

.. autosummary::
   :toctree: _autosummary

    MambuPy.rest.mambustruct
    MambuPy.rest.mambuclient
    MambuPy.rest.mambugroup
    MambuPy.rest.mambuloan
    MambuPy.rest.mamburepayment
    MambuPy.rest.mambutransaction
    MambuPy.rest.mambuproduct
    MambuPy.rest.mambubranch
    MambuPy.rest.mambucentre
    MambuPy.rest.mambuactivity
    MambuPy.rest.mambutask
    MambuPy.rest.mambusaving
    MambuPy.rest.mambusavingtransaction
    MambuPy.rest.mambutransactionchannel
    MambuPy.rest.mambuuser
    MambuPy.rest.mamburoles

.. warning:: This package is soon to be deprecated. Use the :any:`MambuPy.api` package instead.

The REST API way is the currently most developed code on *MambuPy*.

Lives under the :any:`MambuPy.rest` package

It has a lot of objects which model some Mambu entity.

Every object inherits from the parent :any:`mambustruct.MambuStruct`
class. Start at the documentation there for more info on how it works.

.. todo:: Implement objects to make POST requests. The suggestion may be to use
          MambuStruct, to default the __init__ to make a POST request (via the data
          argument) and the attrs attribute to store the elements of the response that
          Mambu gives when a successful POST is achieved.
.. todo:: Implement a lot of other Mambu entities available through GET requests on
          Mambu.
.. todo:: Implement a lot of lacking GET filters on the currently available Mambu
          objects, inside the urlfuncs on the mambuutil module.
.. todo:: A lot of TODO comments are inserted inside the pydocs of the code
          itself. Please read them for suggestions on work need to be done.
"""

from .. import mambuconfig, mambuutil



# coding: utf-8
"""Mambu Loans objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

MambuLoan holds a loan account.

MambuLoans holds a list of loan accounts.

Uses mambugeturl.getloans as default urlfunc.
You can override this and use mambugeturl.getgrouploansurl for the loans
of a specific group instead (you should send the Group ID instead of the
Loan ID as the entid argument in the constructor in that case).
"""


from ..mambugeturl import getloanscustominformationurl, getloansurl, getpostdocumentsurl
from ..mambuutil import MambuError, strip_tags
from .mambustruct import MambuStruct
from .mamburestutils import MambuStructIterator

mod_urlfunc = getloansurl


# Objeto con una Cuenta desde Mambu
class MambuLoan(MambuStruct):
    """A Loan account from Mambu.

    With the default urlfunc, entid argument must be the ID of the
    loan account you wish to retrieve.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """Tasks done here:

        Just initializes the MambuStruct.
        """
        if "getClientDetails" in kwargs and kwargs["getClientDetails"] is not None:
            self.getClientDetails = kwargs["getClientDetails"]
            kwargs.pop("getClientDetails")
        MambuStruct.__init__(
            self, urlfunc, entid, custom_field_name="customFieldValues", *args, **kwargs
        )

    def getDebt(self):
        """Sums up all the balances of the account and returns them."""
        debt = float(self["principalBalance"]) + float(self["interestBalance"])
        debt += float(self["feesBalance"]) + float(self["penaltyBalance"])

        return debt

    def preprocess(self):
        """Preprocessing.

        Each active custom field is given a 'name' key that holds the field
        name, and for each keyed name, the value of the custom field is
        assigned.

        Notes on the group get some html tags removed.
        """
        super(MambuLoan, self).preprocess()

        try:
            self["notes"] = strip_tags(self["notes"])
        except KeyError:
            pass

    def setRepayments(self, *args, **kwargs):
        """Adds the repayments for this loan to a 'repayments' field.

                Repayments are MambuRepayment objects.

                Repayments get sorted by due date.

                Returns the number of requests done to Mambu.

        .. todo:: since pagination logic was added, is not always true that
                  just 1 request was done. It may be more! But since request
                  counter singleton holds true information about how many requests
                  were done to Mambu, in fact this return value may be obsolete
        """

        def duedate(repayment):
            """Util function used for sorting repayments according to due Date"""
            try:
                return repayment["dueDate"]
            except KeyError:
                from datetime import datetime

                return datetime.now()

        try:
            self.mamburepaymentsclass
        except AttributeError:
            from .mamburepayment import MambuRepayments

            self.mamburepaymentsclass = MambuRepayments

        reps = self.mamburepaymentsclass(entid=self["id"], *args, **kwargs)
        reps.attrs = sorted(reps.attrs, key=duedate)
        self["repayments"] = reps

        return 1

    def setTransactions(self, *args, **kwargs):
        """Adds the transactions for this loan to a 'transactions' field.

                Transactions are MambuTransaction objects.

                Transactions get sorted by transaction id.

                Returns the number of requests done to Mambu.

        .. todo:: since pagination logic was added, is not always true that
                  just 1 request was done. It may be more! But since request
                  counter singleton holds true information about how many requests
                  were done to Mambu, in fact this return value may be obsolete
        """

        def transactionid(transaction):
            """Util function used for sorting transactions according to id"""
            try:
                return transaction["transactionId"]
            except KeyError:
                return None

        try:
            self.mambutransactionsclass
        except AttributeError:
            from .mambutransaction import MambuTransactions

            self.mambutransactionsclass = MambuTransactions

        trans = self.mambutransactionsclass(entid=self["id"], *args, **kwargs)
        trans.attrs = sorted(trans.attrs, key=transactionid)
        self["transactions"] = trans

        return 1

    def setBranch(self, *args, **kwargs):
        """Adds the branch for this loan to a 'assignedBranch' field.

        Also adds an 'assignedBranchName' field with the name of the branch.

        Branch is a MambuBranch object.

        Returns the number of requests done to Mambu.
        """

        try:
            self.mambubranchclass
        except AttributeError:
            from .mambubranch import MambuBranch

            self.mambubranchclass = MambuBranch

        branch = self.mambubranchclass(entid=self["assignedBranchKey"], *args, **kwargs)
        self["assignedBranchName"] = branch["name"]
        self["assignedBranch"] = branch

        return 1

    def setCentre(self, *args, **kwargs):
        """Adds the centre for this loan to a 'assignedCentre' field.

        Also adds an 'assignedCentreName' field with the name of the centre.

        Centre is a MambuCentre object.

        Returns the number of requests done to Mambu.
        """

        try:
            self.mambucentreclass
        except AttributeError:
            from .mambucentre import MambuCentre

            self.mambucentreclass = MambuCentre

        centre = self.mambucentreclass(entid=self["assignedCentreKey"], *args, **kwargs)

        self["assignedCentreName"] = centre["name"]
        self["assignedCentre"] = centre

        return 1

    def setUser(self, *args, **kwargs):
        """Adds the user for this loan to a 'user' field.

        User is a MambuUser object.

        Returns the number of requests done to Mambu.
        """
        try:
            self.mambuuserclass
        except AttributeError:
            from .mambuuser import MambuUser

            self.mambuuserclass = MambuUser

        try:
            user = self.mambuuserclass(entid=self["assignedUserKey"], *args, **kwargs)
        except KeyError:
            err = MambuError("Loan Account %s has no assigned user" % self["id"])
            err.noUser = True
            raise err

        self["user"] = user

        return 1

    def setProduct(self, *args, **kwargs):
        """Adds the product for this loan to a 'product' field.

        Product is a MambuProduct object.

        cache argument allows to use AllMambuProducts singleton to
        retrieve the products. See mambuproduct.AllMambuProducts code
        and pydoc for further information.

        Returns the number of requests done to Mambu.
        """
        cache = kwargs.get("cache", False)
        if cache:
            try:
                self.allmambuproductsclass
            except AttributeError:
                from .mambuproduct import AllMambuProducts

                self.allmambuproductsclass = AllMambuProducts

            prods = self.allmambuproductsclass(*args, **kwargs)
            for prod in prods:
                if prod["encodedKey"] == self["productTypeKey"]:
                    self["product"] = prod
            try:
                # asked for cache, but cache was originally empty
                prods.noinit
            except AttributeError:
                return 1
            return 0
        try:
            self.mambuproductclass
        except AttributeError:
            from .mambuproduct import MambuProduct

            self.mambuproductclass = MambuProduct

        product = self.mambuproductclass(entid=self["productTypeKey"], *args, **kwargs)

        self["product"] = product

        return 1

    def _get_roles(self, fullDetails=True, *args, **kwargs):
        """ Get a list of roles and clients for the loan account.

        The roles are mentionend in the holder of the loan account.

        The holder acquires a "roles" property, a list of dictionaries
        with the following keys:

          role (str): role name
          client (obj): MambuClient that has the mentioned role

        Args:
          fullDetails (bool): whether instantiate clients with
                              full details or not

        Returns:
          int - number of requests done to Mambu
        """
        requests = 0
        self["holder"]["roles"] = []
        # If holder is group, attach role client data to the group
        for c in self["holder"]["groupRoles"]:
            try:
                self.mambuclientclass
            except AttributeError:
                from .mambuclient import MambuClient
                self.mambuclientclass = MambuClient

            cli = self.mambuclientclass(
                entid=c["clientKey"], fullDetails=fullDetails, *args, **kwargs
            )
            self["holder"]["roles"].append(
                {"role": c["roleName"], "client": cli})
            requests += 1

        return requests

    def _get_loanclientsdata_from_group_setHolder(self, holder, *args, **kwargs):
        loanclients = {}

        loanclientsdata = self.getClientDetails(holder=holder, *args, **kwargs)

        for data in loanclientsdata:
            loanclients[data["id"]] = {
                "client": data["client"],
                "name": data["name"],
                "loan": self,
                "amount": data["amount"],
                "montoPago": data["amount"]
                / float(self["repaymentInstallments"]),
                "porcentaje": data["amount"] / float(self["loanAmount"]),
            }
            # Any extra key,val pair on loannames is plainly assigned to the loanclients[cte] dict
            for k, v in [
                (key, val)
                for (key, val) in data.items()
                if key not in ["amount", "name"]
            ]:
                loanclients[data["id"]][k] = v

        return loanclients

    def _get_holder_data_by_group(self, getRoles, fullDetails, getClients, *args, **kwargs):
        requests = 0
        self["holderType"] = "Grupo"
        try:
            self.mambugroupclass
        except AttributeError:
            from .mambugroup import MambuGroup

            self.mambugroupclass = MambuGroup

        holder = self.mambugroupclass(
            entid=self["accountHolderKey"], fullDetails=True, *args, **kwargs
        )
        self["holder"] = holder
        requests += 1

        if getRoles:
            requests += self._get_roles(fullDetails, *args, **kwargs)

        if getClients:
            try:
                self.mambuclientclass
            except AttributeError:  # pragma: no cover
                from .mambuclient import MambuClient
                self.mambuclientclass = MambuClient
            requests += holder.setClients(
                fullDetails=fullDetails,
                mambuclientclass=self.mambuclientclass,
                *args,
                **kwargs
            )

            self["clients"] = self._get_loanclientsdata_from_group_setHolder(holder, *args, **kwargs)

        return requests

    def _get_holder_data_by_client(self, fullDetails, getClients, *args, **kwargs):
        requests = 0
        self["holderType"] = "Cliente"
        try:
            self.mambuclientclass
        except AttributeError:
            from .mambuclient import MambuClient

            self.mambuclientclass = MambuClient

        holder = self.mambuclientclass(
            entid=self["accountHolderKey"], fullDetails=fullDetails, *args, **kwargs
        )
        self["holder"] = holder
        requests += 1

        if getClients:
            monto = float(self["loanAmount"])
            self["clients"] = {
                holder["name"]: {
                    "client": holder,
                    "loan": self,
                    "amount": monto,
                    "montoPago": monto / float(self["repaymentInstallments"]),
                    "porcentaje": 1.0,
                }
            }

        return requests

    def setHolder(self, getClients=False, getRoles=False, *args, **kwargs):
        """Adds the "holder" of the loan to a 'holder' field.

                Holder may be a MambuClient or a MambuGroup object, depending on
                the type of loan created at Mambu.

                getRoles argument tells this method to retrive the clients who
                have specific roles in the group, in case of a group holder.

                getClients argument tells this method to retrieve the client
                members of the group, in case of a group holder. See the
                definition of the 'clients' field added with this argument
                below.

                fullDetails argument works for Client Holder and for the client
                members of the group. But the Group is always retrieved with
                full details.

                When using getRoles argument also adds a 'roles' field to the
                holder, a dictionary with keys:

                * 'role' , the rolename

                * 'client', a MambuClient having that role on the group

        .. todo:: since the roles field is added to the holder, and it only
                  applies to Groups, perhaps the getRoles functionality should be
                  on MambuGroup code

                When using getClients argument, also adds a 'clients' field to
                the loan. By the way it works, when holder is a Group, it calls
                the setClients method, which adds a 'clients' field to the
                MambuGroup.

                The 'clients' field on the loan is a dictionary with the full
                name of each client receiving from this loan account as a key
                (see the 'name' field on MambuClient and MambuGroup). Each key
                holds another dictionary with the following keys:

                * 'client' holds a MambuClient. So even if a group is the holder
                  of the account, you can access each individual client object
                  here.

                * 'loan' holds a reference to this loan account (self)

                * 'amount' defaults to the total loan amount ('loanAmount'
                  field) (see the pydoc for getClientDetails method)

                * 'montoPago' defaults to the total loan amount divided by the
                  number of repayment installments of the loan account (the
                  'repaymentInstallments' field), that is, how much to pay for
                  each repayment installment (see the pydoc for getClientDetails
                  method)

        .. todo:: change the name to an english name

                * 'porcentaje' defaults to a 100%, the total loan amount divided
                  by itself, that is, hoy much in percentage of the debt belongs
                  to this client (see the pydoc for getClientDetails method)

        .. todo:: change the name to an english name

                This three fields are supposed to reflect individual numbers for
                each client, but Mambu does an awful job indicating this on
                group loan accounts. So you should indicate how much you
                assigned to each client in other ways. Using the notes of the
                loan account perhaps (with a predefined format you parse on the
                preprocessing of MambuLoan)? with custom fields? Your call that
                you may define making your own loan account class inheriting
                from MambuLoan class. See the pydoc for getClientDetails method
                below, it specifies the way to determine the amount for each
                client depending on your use of Mambu, and additionally any
                other fields you wish to add to the 'clients' field on the loan
                account.

                Returns the number of requests done to Mambu.

        .. todo:: what to do on Hybrid loan accounts?
        """
        if "fullDetails" in kwargs:
            fullDetails = kwargs["fullDetails"]
            kwargs.pop("fullDetails")
        else:
            fullDetails = True

        if self["accountHolderType"] == "GROUP":
            requests = self._get_holder_data_by_group(getRoles, fullDetails, getClients, *args, **kwargs)
        else:  # "CLIENT"
            requests = self._get_holder_data_by_client(fullDetails, getClients, *args, **kwargs)

        return requests

    def getClientDetails(self, *args, **kwargs):
        """Gets the loan details for every client holder of the account.

        As default, assigns the whole loan amount to each client. This
        works fine for Client holders of the loan account. When Group
        holders, this is perhaps not ideal, but I cannot tell.

        If you inherit MambuLoan you should override this method to
        determine another way to assign particular amounts to each
        client.

        You can also use the overriden version of this method to add
        several other fields with information you wish to associate to
        each client holder of the loan account.

        BEWARE: for group loan accounts, this code assumes the holder
        (the group) currentrly has all the client members. But for
        accounts where the holder has changed along time, you may
        stumble upon accounts which assumes certain group members which
        weren't the members that belonged to the group when it was
        disbursed.
        """
        loannames = []

        holder = kwargs["holder"]
        for client in holder["clients"]:
            loannames.append(
                {
                    "id": client["id"],
                    "name": client["name"],
                    "client": client,
                    "amount": self["loanAmount"],
                }
            )

        return loannames

    def setActivities(self, *args, **kwargs):
        """Adds the activities for this loan to a 'activities' field.

        Activities are MambuActivity objects.

        Activities get sorted by activity timestamp.

        Returns the number of requests done to Mambu.
        """

        def activity_date(activity):
            """Util function used for sorting activities according to timestamp"""
            try:
                return activity["activity"]["timestamp"]
            except KeyError:
                return None

        try:
            self.mambuactivitiesclass
        except AttributeError:
            from .mambuactivity import MambuActivities

            self.mambuactivitiesclass = MambuActivities

        activities = self.mambuactivitiesclass(
            loanAccountId=self["encodedKey"], *args, **kwargs
        )
        activities.attrs = sorted(activities.attrs, key=activity_date)
        self["activities"] = activities

        return 1

    def update(self, data, *args, **kwargs):
        """Updates a loan in Mambu

        Updates customFields of a MambuLoan

        TODO: update "core fields" of a MambuLoan

        https://support.mambu.com/docs/loans-api#post-loans
        https://support.mambu.com/docs/loans-api#patch-loan
        https://support.mambu.com/docs/loans-api#patch-loan-custom-field-values

        Parameters
        -data       dictionary with data to update
        """
        cont_requests = 0
        data2update = {}

        # UPDATE customFields
        if data.get("customInformation"):
            data2update = {}
            data2update["customInformation"] = data.get("customInformation")
            self._MambuStruct__urlfunc = getloanscustominformationurl
            cont_requests += self.update_patch(data2update, *args, **kwargs)
            self._MambuStruct__urlfunc = getloansurl

        cont_requests += super(MambuLoan, self).update(data, *args, **kwargs)

        return cont_requests

    def update_patch(self, data, *args, **kwargs):
        """Updates a Mambu loan using method PATCH

        Args:
            data (dictionary): dictionary with data to update

        https://support.mambu.com/docs/loans-api#patch-loan
        """
        return super(MambuLoan, self).update_patch(data, *args, **kwargs)

    def upload_document(self, data, *args, **kwargs):
        """Updates a loan in Mambu

        Uploads a document to a MambuLoan

        https://support.mambu.com/docs/attachments-api#post-attachments

        Parameters
        -data       dictionary with data to upload

        Example
        data = {
            "document":{
                "documentHolderKey"     : self.encodedKey,
                "documentHolderType"    : "LOAN_ACCOUNT",
                "name"                  : "loan_resume",
                "type"                  : "pdf",
            },
            "documentContent"           : "['encodedBase64_file']",
        }
        """
        cont_requests = 0
        # upload document
        self._MambuStruct__urlfunc = getpostdocumentsurl
        cont_requests += super(MambuLoan, self).upload_document(data, *args, **kwargs)
        self._MambuStruct__urlfunc = getloansurl

        return cont_requests


class MambuLoans(MambuStruct):
    """A list of Loan accounts from Mambu.

    With the default urlfunc, entid argument must be empty at
    instantiation time to retrieve all the loan accounts according to
    any other filter you send to the urlfunc.

    There's another possible urlfunc you may use, getgrouploansurl. If
    you use that, you should send the ID of the group from which you
    wish to retrieve its loan accounts.

    itemclass argument allows you to pass some other class as the
    elements for the list. Why is this useful? at least for now,
    MambuLoan is the most specialized class on MambuPy. So you may
    wish to override several behaviours by creating your own
    MambuLoan son class. Pass that to the itemclass argument here
    and voila, you get a list of YourMambuLoan class using
    MambuLoans instead of plain old MambuLoan elements.

    If you wish to specialize other Mambu objects on MambuPy you may
    do that. Mind that if you desire that the iterable version of it
    to have elements of your specialized class, you need to change
    the logic of the constructor and the convert_dict_to_attrs method in
    the iterable class to use some sort of itemclass there too.
    Don't forget to submit the change on a pull request when done
    ;-)
    """

    def __init__(
        self, urlfunc=mod_urlfunc, entid="", itemclass=MambuLoan, *args, **kwargs
    ):
        """By default, entid argument is empty. That makes perfect
        sense: you want several groups, not just one.
        """
        if "getClientDetails" in kwargs and kwargs["getClientDetails"] is not None:
            self.getClientDetails = kwargs["getClientDetails"]
            kwargs.pop("getClientDetails")
        self.itemclass = itemclass
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """The trick for iterable Mambu Objects comes here:

                You iterate over each element of the responded List from Mambu,
                and create a Mambu Loan (or your own itemclass) object for each
                one, initializing them one at a time, and changing the attrs
                attribute (which just holds a list of plain dictionaries) with a
                MambuLoan (or your own itemclass) just created.

        .. todo:: pass a valid (perhaps default) urlfunc, and its
                  corresponding id to entid to each itemclass, telling MambuStruct
                  not to connect() by default. It's desirable to connect at any
                  other further moment to refresh some element in the list.
        """
        for n, l in enumerate(self.attrs):
            # ok ok, I'm modifying elements of a list while iterating it. BAD PRACTICE!
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            try:
                self.mambuloanclass
            except AttributeError:
                self.mambuloanclass = self.itemclass

            loan = self.mambuloanclass(urlfunc=None, entid=None, *args, **kwargs)
            loan.init(l, *args, **kwargs)
            loan._MambuStruct__urlfunc = getloansurl
            self.attrs[n] = loan



# coding: utf-8
"""Mambu Products objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

MambuProduct holds a product.

MambuProducts holds a list of products.

Uses mambugeturl.getproducturl as default urlfunc
"""


from ..mambugeturl import getproductsurl
from .mambustruct import MambuStruct
from .mamburestutils import MambuStructIterator

mod_urlfunc = getproductsurl


class MambuProduct(MambuStruct):
    """A Product from Mambu.

    With the default urlfunc, entid argument must be the ID of the
    product you wish to retrieve.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """Tasks done here:

        Just initializes the MambuStruct.
        """
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)


class MambuProducts(MambuStruct):
    """A list of Products from Mambu.

    With the default urlfunc, entid argument must be empty at
    instantiation time to retrieve all the products according to any
    other filter you send to the urlfunc.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """By default, entid argument is empty. That makes perfect
        sense: you want several products, not just one
        """
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """The trick for iterable Mambu Objects comes here:

                You iterate over each element of the responded List from Mambu,
                and create a Mambu Product object for each one, initializing them
                one at a time, and changing the attrs attribute (which just
                holds a list of plain dictionaries) with a MambuProduct just
                created.

        .. todo:: pass a valid (perhaps default) urlfunc, and its
                  corresponding id to entid to each MambuProduct, telling
                  MambuStruct not to connect() by default. It's desirable to
                  connect at any other further moment to refresh some element in
                  the list.
        """
        for n, r in enumerate(self.attrs):
            # ok ok, I'm modifying elements of a list while iterating it. BAD PRACTICE!
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            try:
                self.mambuproductclass
            except AttributeError:
                self.mambuproductclass = MambuProduct

            product = self.mambuproductclass(urlfunc=None, entid=None, *args, **kwargs)
            product.init(r, *args, **kwargs)
            self.attrs[n] = product


class AllMambuProducts(MambuStruct):
    """Singleton that holds ALL the Mambu products.

        With the default urlfunc, entid argument must be empty at
        instantiation time to retrieve all the products according to any
        other filter you send to the urlfunc.

        Caching singleton. You may not wish to retrieve from Mambu all the
        products every time you need to use them, so you can use this, which
        requests from Mambu one time only, and holds them here forever
        during your python session.

        To use it, instead of instantiating a MambuProducts() object when
        needing all the products from Mambu, instantiate an
        AllMambuProducts() object and the caching will be used by default.

    .. todo:: is there a better way to implement cache directly on
              MambuProducts()?

        Why have a cache? because Mambu products are not likely to change
        once you have them configured in Mambu. So it's better to prevent a
        lot of requests to get the same information over and over again.

        You are free to implement caches such as this on any other Mambu
        objects if you find it useful (perhaps groups or branches or any
        other Mambu entity do not change a lot on your own business?).
        Please try to be consistent on names and functionality.
    """

    __instance = None

    def __new__(cls, *args, **kwargs):
        if not cls.__instance:
            # cls.__instance = super(AllMambuProducts, cls).__new__(cls, *args, **kwargs)
            cls.__instance = super(AllMambuProducts, cls).__new__(cls)
        else:
            cls.__instance.noinit = True
        return cls.__instance

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """If you have already retrieved products, you don't contact
        Mambu. If you haven't, you do.
        """
        try:
            getattr(self, "noinit")
        except AttributeError:
            MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def __getattribute__(self, name):
        """Object-like get attribute"""
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            # Iterable AllMambuProducts singleton also uses a special
            # noinit property that should raise AttributeError when
            # not set
            if name in ["params", "noinit", "mambuproductclass"]:
                raise AttributeError
            return self[name]

    def convert_dict_to_attrs(self, *args, **kwargs):
        """The trick for iterable Mambu Objects comes here:

                You iterate over each element of the responded List from Mambu,
                and create a Mambu Product object for each one, initializing them
                one at a time, and changing the attrs attribute (which just
                holds a list of plain dictionaries) with a MambuProduct just
                created.

        .. todo:: pass a valid (perhaps default) urlfunc, and its
                  corresponding id to entid to each MambuProduct, telling
                  MambuStruct not to connect() by default. It's desirable to
                  connect at any other further moment to refresh some element in
                  the list.
        """
        for n, r in enumerate(self.attrs):
            # ok ok, I'm modifying elements of a list while iterating it. BAD PRACTICE!
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            try:
                self.mambuproductclass
            except AttributeError:
                self.mambuproductclass = MambuProduct

            product = self.mambuproductclass(urlfunc=None, entid=None, *args, **kwargs)
            product.init(r, *args, **kwargs)
            product._MambuStruct__urlfunc = getproductsurl
            self.attrs[n] = product



# coding: utf-8
"""Mambu Users objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

MambuUser holds a user.

MambuUsers holds a list of users.

Uses mambugeturl.getuserurl as default urlfunc
"""


from ..mambugeturl import getusercustominformationurl, getuserurl
from .mambustruct import MambuStruct
from .mamburestutils import MambuStructIterator

mod_urlfunc = getuserurl


class MambuUser(MambuStruct):
    """A User from Mambu.

    With the default urlfunc, entid argument must be the ID of the
    user you wish to retrieve.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """Tasks done here:

        Just initializes the MambuStruct.
        """
        MambuStruct.__init__(
            self, urlfunc, entid, custom_field_name="customFields", *args, **kwargs
        )

    def preprocess(self):
        """Preprocessing.

        Removes repeated chars from firstName and lastName fields.

        Adds a 'name' field joining all names in to one string.
        """
        super(MambuUser, self).preprocess()

        try:
            self["firstName"] = self["firstName"].strip()
        except Exception:
            self["firstName"] = ""
        try:
            self["lastName"] = self["lastName"].strip()
        except Exception:
            self["lastName"] = ""

        self["name"] = self["firstName"] + " " + self["lastName"]

    def setGroups(self, *args, **kwargs):
        """Adds the groups assigned to this user to a 'groups' field.

        Returns the number of requests done to Mambu.
        """
        try:
            self.mambugroupsclass
        except AttributeError:
            from .mambugroup import MambuGroups

            self.mambugroupsclass = MambuGroups

        groups = self.mambugroupsclass(
            creditOfficerUsername=self["username"], *args, **kwargs
        )
        self["groups"] = groups

        return 1

    def setRoles(self, *args, **kwargs):
        """Adds the role assigned to this user to a 'role' field.

        Depends on the 'role' field that comes with a fullDetails=True
        build of the MambuUser.

        Returns the number of requests done to Mambu.
        """
        try:
            self.mamburoleclass
        except AttributeError:
            from .mamburoles import MambuRole

            self.mamburoleclass = MambuRole

        try:
            role = self.mamburoleclass(entid=self["role"]["encodedKey"], *args, **kwargs)
        except KeyError:
            return 0
        self["role"]["role"] = role

        return 1

    def setBranch(self, *args, **kwargs):
        """Adds the branch to which this user belongs"""
        try:
            self.mambubranchclass
        except AttributeError:
            from .mambubranch import MambuBranch

            self.mambubranchclass = MambuBranch
        try:
            branch = self.mambubranchclass(
                entid=self["assignedBranchKey"], *args, **kwargs
            )
            self["branch"] = branch
        except KeyError:
            self["branch"] = ""
        return 1

    def create(self, data, *args, **kwargs):
        """Creates an user in Mambu

        Parameters
        -data       dictionary with data to send

        """
        super(MambuUser, self).create(data)
        self["user"][self.custom_field_name] = self["customInformation"]
        self.init(attrs=self["user"])
        return 1

    def update(self, data, *args, **kwargs):
        """Updates a user in Mambu

        Updates customFields of a MambuUser

        https://support.mambu.com/docs/users-api#post-users
        https://support.mambu.com/docs/users-api#parameters-for-patch-custom-fields-method-for-user

        Parameters
        -data       dictionary with data to update
        """
        cont_requests = 0
        data2update = {}

        # UPDATE customFields
        if data.get("customInformation"):
            data2update = {"user": {}}
            data2update["customInformation"] = data.get("customInformation")
            self._MambuStruct__urlfunc = getusercustominformationurl
            cont_requests += self.update_patch(data2update, *args, **kwargs)
            self._MambuStruct__urlfunc = getuserurl

        cont_requests += super(MambuUser, self).update(data, *args, **kwargs)
        return cont_requests

    def update_patch(self, data, *args, **kwargs):
        """Updates a Mambu user using method PATCH

        Args:
            data (dictionary): dictionary with data to update

        https://support.mambu.com/docs/users-api#patch-user-custom-field-values
        """
        return super(MambuUser, self).update_patch(data, *args, **kwargs)


class MambuUsers(MambuStruct):
    """A list of Users from Mambu.

    With the default urlfunc, entid argument must be empty at
    instantiation time to retrieve all the users according to any other
    filter you send to the urlfunc.

    itemclass argument allows you to pass some other class as the
    elements for the list. Why is this useful? You may wish to override
    several behaviours by creating your own MambuUser son class. Pass
    that to the itemclass argument here and voila, you get a list of
    YourMambuUser class using MambuUsers instead of plain old MambuUser
    elements.

    If you wish to specialize other Mambu objects on MambuPy you may
    do that. Mind that if you desire that the iterable version of it
    to have elements of your specialized class, you need to change
    the logic of the constructor and the convert_dict_to_attrs method in
    the iterable class to use some sort of itemclass there too.
    Don't forget to submit the change on a pull request when done
    ;-)
    """

    def __init__(
        self, urlfunc=mod_urlfunc, entid="", itemclass=MambuUser, *args, **kwargs
    ):
        """By default, entid argument is empty. That makes perfect
        sense: you want several branches, not just one
        """
        self.itemclass = itemclass
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """The trick for iterable Mambu Objects comes here:

                You iterate over each element of the responded List from Mambu,
                and create a Mambu User (or your own itemclass) object for each
                one, initializing them one at a time, and changing the attrs
                attribute (which just holds a list of plain dictionaries) with a
                MambuUser (or your own itemclass) just created.

        .. todo:: pass a valid (perhaps default) urlfunc, and its
                  corresponding id to entid to each itemclass, telling MambuStruct
                  not to connect() by default. It's desirable to connect at any
                  other further moment to refresh some element in the list.
        """
        for n, u in enumerate(self.attrs):
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            try:
                self.mambuuserclass
            except AttributeError:
                self.mambuuserclass = self.itemclass

            user = self.mambuuserclass(urlfunc=None, entid=None, *args, **kwargs)
            user.init(u, *args, **kwargs)
            user._MambuStruct__urlfunc = getuserurl
            self.attrs[n] = user



# coding: utf-8
"""Mambu Transaction Channel objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

MambuTansactionChannel holds a transaction channel.

MambuTansactionChannels holds a list of transaction channels.

Uses mambugeturl.gettransactionchannelsurl as default urlfunc.
"""


from ..mambugeturl import gettransactionchannelsurl
from .mambustruct import MambuStruct
from .mamburestutils import MambuStructIterator

mod_urlfunc = gettransactionchannelsurl


# Objeto con una Cuenta desde Mambu
class MambuTransactionChannel(MambuStruct):
    """A Transaction Channel from Mambu.

    With the default urlfunc, entid argument must be the ID of the
    transaction channel you wish to retrieve.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """Tasks done here:

        Just initializes the MambuStruct.
        """
        MambuStruct.__init__(
            self, urlfunc, entid, custom_field_name="customFieldValues", *args, **kwargs
        )

    def preprocess(self):
        """Preprocessing.

        Each active custom field is given a 'name' key that holds the field
        name, and for each keyed name, the value of the custom field is
        assigned.
        """
        super(MambuTransactionChannel, self).preprocess()


class MambuTransactionChannels(MambuStruct):
    """A list of Transaction Channels from Mambu.

    With the default urlfunc, entid argument must be empty at
    instantiation time to retrieve all the transaction channels according to
    any other filter you send to the urlfunc.
    """

    def __init__(
        self,
        urlfunc=mod_urlfunc,
        entid="",
        itemclass=MambuTransactionChannel,
        *args,
        **kwargs
    ):
        """By default, entid argument is empty. That makes perfect
        sense: you want several groups, not just one.
        """
        self.itemclass = itemclass
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """The trick for iterable Mambu Objects comes here:

        You iterate over each element of the responded List from Mambu,
        and create a TransactionChannel (or your own itemclass) object for each
        one, initializing them one at a time, and changing the attrs
        attribute (which just holds a list of plain dictionaries) with a
        TransactionChannel (or your own itemclass) just created.

        .. todo:: pass a valid (perhaps default) urlfunc, and its
          corresponding id to entid to each itemclass, telling MambuStruct
          not to connect() by default. It's desirable to connect at any
          other further moment to refresh some element in the list.
        """
        for n, l in enumerate(self.attrs):
            # ok ok, I'm modifying elements of a list while iterating it. BAD PRACTICE!
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            transaction_channel = self.itemclass(
                urlfunc=None, entid=None, *args, **kwargs
            )
            transaction_channel.init(l, *args, **kwargs)
            transaction_channel._MambuStruct__urlfunc = gettransactionchannelsurl
            self.attrs[n] = transaction_channel



# coding: utf-8
"""Mambu Task objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

MambuTask holds a task. Look at MambuTask pydoc for further info.

Uses mambugeturl.gettasksurl as default urlfunc
"""
from ..mambugeturl import gettasksurl
from .mambustruct import MambuStruct
from .mamburestutils import MambuStructIterator

mod_urlfunc = gettasksurl


class MambuTask(MambuStruct):
    """A Task from Mambu.

    Don't instantiate this class directly. It's mostly used by
    MambuTasks to configure each of its elements as MambuTask
    objects. There's no suitable urlfunc to use to retrieve just a
    specific transaction from a loan account. In fact, you can look at
    the code of MambuTasks.convert_dict_to_attrs(), it uses urlfunc
    and entid = None , so no connection to Mambu will be made, never,
    for any particular MambuTask object.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """Tasks done here:

        Just initializes the MambuStruct.
        """
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __repr__(self):
        """Instead of the default id given by the parent class, shows
        the id, title, dueDate, status of the task.
        """
        try:
            return self.__class__.__name__ + " - taskid: '%s', %s, %s" % (
                self["task"]["id"],
                self["task"]["dueDate"],
                self["task"]["status"],
            )
        except KeyError:
            try:
                return self.__class__.__name__ + " - taskid: '%s'" % self["task"]["id"]
            except KeyError:
                try:
                    return self.__class__.__name__ + " - taskid: '%s', %s, %s" % (
                        self["id"],
                        self["dueDate"],
                        self["status"],
                    )
                except KeyError:
                    return self.__class__.__name__ + " - taskid: '%s'" % self["id"]

    def preprocess(self):
        """Preprocessing.

        Flattens the object. When .creat[ed]() the data comes on the
        'task' dictionary inside of the response. Instead, every
        element of the 'task' dictionary is taken out to the main
        attrs dictionary, since GET response comes in that format.
        """
        super(MambuTask, self).preprocess()

        try:
            for k, v in self["task"].items():
                self[k] = v
            del self.attrs["task"]
        except Exception:
            pass

    def create(self, data, *args, **kwargs):
        """Create a task in Mambu

        Parameters
        -data       dictionary with data to send

        """
        return super(MambuTask, self).create(data)

    def close(self, *args, **kwargs):
        """"""
        from datetime import datetime

        data = {
            "task": {
                "status": "COMPLETED",
                "completionDate": datetime.now().strftime("%Y-%m-%d"),
                "encodedKey": self["encodedKey"],
                "dueDate": self["dueDate"].strftime("%Y-%m-%d"),
                "title": self["title"],
                "description": self["description"],
                "assignedUserKey": self["assignedUserKey"],
            }
        }
        try:
            data["task"]["taskLinkType"] = self["taskLinkType"]
        except KeyError:
            pass
        try:
            data["task"]["taskLinkKey"] = self["taskLinkKey"]
        except KeyError:
            pass

        self._MambuStruct__method = "POST"
        self._MambuStruct__urlfunc = gettasksurl
        self._MambuStruct__data = data
        self.connect(*args, **kwargs)

        data["task"]["id"] = self["id"]
        data["task"]["completionDate"] = datetime.strptime(
            data["task"]["completionDate"], "%Y-%m-%d"
        )
        data["task"]["dueDate"] = datetime.strptime(data["task"]["dueDate"], "%Y-%m-%d")
        self.attrs = data["task"]
        self.convert_dict_to_attrs()

        self._MambuStruct__method = "GET"
        self._MambuStruct__urlfunc = None
        self._MambuStruct__data = None


class MambuTasks(MambuStruct):
    """A list of Tasks from Mambu."""

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """The trick for iterable Mambu Objects comes here:

        You iterate over each element of the responded List from Mambu,
        and create a Mambu Task object for each one, initializing
        them one at a time, and changing the attrs attribute (which just
        holds a list of plain dictionaries) with a MambuTask just
        created.

        """
        for n, a in enumerate(self.attrs):
            # ok ok, I'm modifying elements of a list while iterating it. BAD PRACTICE!
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            try:
                self.mambutaskclass
            except AttributeError:
                self.mambutaskclass = MambuTask

            task = self.mambutaskclass(urlfunc=None, entid=None, *args, **kwargs)
            task.init(a, *args, **kwargs)
            task._MambuStruct__urlfunc = gettasksurl
            self.attrs[n] = task



# coding: utf-8
"""Mambu Activity objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

MambuActivity holds a mambu activity. Don't instantiate this class
directly. Look at MambuActivity pydoc for further info.

MambuActivities holds a list of activities.

Uses mambugeturl.getactivitiesurl as default urlfunc
"""


from ..mambugeturl import getactivitiesurl
from .mambustruct import MambuStruct
from .mamburestutils import MambuStructIterator

mod_urlfunc = getactivitiesurl


class MambuActivity(MambuStruct):
    """A loan Activity from Mambu.

    Don't instantiate this class directly. It's mostly used by
    MambuActivities to configure each of its elements as
    MambuActivity objects. There's no suitable urlfunc to use to
    retrieve just a specific transaction from a loan account. In fact,
    you can look at the code of MambuActivities.convert_dict_to_attrs(),
    it uses urlfunc and entid = None , so no connection to Mambu will be
    made, never, for any particular MambuActivity object.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """Tasks done here:

        Just initializes the MambuStruct.
        """
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __repr__(self):
        """Instead of the default id given by the parent class, shows
        the transactionid of the transaction.
        """
        try:
            return self.__class__.__name__ + " - activityid: %s" % self["activity"]
        except KeyError:
            return self.__class__.__name__ + " - activityid: %s" % self["activity"]


class MambuActivities(MambuStruct):
    """A list of loan Activities from Mambu."""

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """The trick for iterable Mambu Objects comes here:

        You iterate over each element of the responded List from Mambu,
        and create a Mambu Activity object for each one, initializing
        them one at a time, and changing the attrs attribute (which just
        holds a list of plain dictionaries) with a MambuActivity just
        created.

        """
        for n, a in enumerate(self.attrs):
            # ok ok, I'm modifying elements of a list while iterating it. BAD PRACTICE!
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            try:
                self.mambuactivityclass
            except AttributeError:
                self.mambuactivityclass = MambuActivity

            activity = self.mambuactivityclass(urlfunc=None, entid=None, *args, **kwargs)
            activity.init(a, *args, **kwargs)
            activity._MambuStruct__urlfunc = getactivitiesurl
            self.attrs[n] = activity



# coding: utf-8
"""Mambu Clients objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

MambuClient holds a client.

MambuClients holds a list of clients.

Uses mambugeturl.getclienturl as default urlfunc
"""


unicode = str

from ..mambugeturl import getclientcustominformationurl, getclienturl
from ..mambuutil import strip_consecutive_repeated_char as scrc
from .mambustruct import MambuStruct
from .mamburestutils import MambuStructIterator

mod_urlfunc = getclienturl


class MambuClient(MambuStruct):
    """A Client from Mambu.

    With the default urlfunc, entid argument must be the ID of the
    client you wish to retrieve.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """Tasks done here:

        Just initializes the MambuStruct.
        """
        MambuStruct.__init__(
            self, urlfunc, entid, custom_field_name="customInformation", *args, **kwargs
        )

    def preprocess(self):
        """Preprocessing.

        Flattens the object. Important data comes on the 'client'
        dictionary inside of the response. Instead, every element of the
        'client' dictionary is taken out to the main attrs dictionary.

        Removes repeated chars from firstName, middleName and lastName
        fields. Adds a firstLastName and secondLastName fields for
        clients using more than one last name.

        Adds a 'name' field joining all names in to one string. IMHO
        Mambu does an awful job with names. Besides, this is useful on
        loan accounts. They get a 'holder' field added somewhere, which
        may be an individual client, or a group. So to get the holder
        name, you just access the 'holder'['name']. No matter if you
        have a client loan or a group loan, you get the name of the
        holder.

        Creates and 'address' field that holds the very first address on
        the addresses field, to ease the use of the
        perhaps-current-or-most-important address of the client.

        Creates a field for each documentId type with the value of the
        documentId, to ease the use of docuemnt IDs. That way if you
        have a document Id named 'CURP', then you access it directly via
        the 'CURP' field, instead of going all the way through the
        idDocuments field.
        """
        super(MambuClient, self).preprocess()

        try:
            for k, v in self["client"].items():
                self[k] = v
            del self.attrs["client"]
        except Exception:
            pass

        try:
            self["firstName"] = scrc(self["firstName"], " ").strip()
        except Exception:
            self["firstName"] = ""
        try:
            self["middleName"] = scrc(self["middleName"], " ").strip()
        except Exception:
            self["middleName"] = ""
        self["givenName"] = scrc(
            self["firstName"]
            + ((" " + self["middleName"]) if self["middleName"] != "" else ""),
            " ",
        ).strip()
        self["lastName"] = scrc(self["lastName"], " ").strip()
        self["firstLastName"] = (
            " ".join(self["lastName"].split(" ")[:-1])
            if len(self["lastName"].split(" ")) > 1
            else self["lastName"]
        )
        self["secondLastName"] = (
            " ".join(self["lastName"].split(" ")[-1:])
            if len(self["lastName"].split(" ")) > 1
            else ""
        )

        self["name"] = scrc("%s %s" % (self["givenName"], self["lastName"]), " ").strip()

        self["address"] = {}
        try:
            for name, item in self["addresses"][0].items():
                try:
                    self["addresses"][0][name] = item.strip()
                    self["address"][name] = item.strip()
                except AttributeError:
                    pass
        except (KeyError, IndexError):
            pass

        try:
            for idDoc in self["idDocuments"]:
                self[idDoc["documentType"]] = idDoc["documentId"]
        except KeyError:
            pass

    def postprocess(self):
        """Postprocessing.

                Just in case some elements on the addresses was converted to
                anything but string, it gets converted back to only string
                (unicode). Things on addresses are not useful but by what they
                say, not what they are.

        .. todo:: do the same thing to the 'address' field created on
                  preprocessing.
        """
        try:
            for name, item in self["addresses"][0].items():
                try:
                    if name == "indexInList":
                        continue
                    self["addresses"][0][name] = unicode(self["addresses"][0][name])
                    self["address"][name] = unicode(self["address"][name])
                except AttributeError:
                    pass
        except (KeyError, IndexError):
            pass

        super(MambuClient, self).postprocess()

    def setGroups(self, *args, **kwargs):
        """Adds the groups to which this client belongs.

        The 'groupKeys' field of the client holds a list of the
        encodedKeys of the groups to which this client belongs.

        Returns the number of requests done to Mambu.
        """

        requests = 0
        groups = []
        try:
            for gk in self["groupKeys"]:
                try:
                    self.mambugroupclass
                except AttributeError:
                    from .mambugroup import MambuGroup

                    self.mambugroupclass = MambuGroup

                g = self.mambugroupclass(entid=gk, *args, **kwargs)
                requests += 1
                groups.append(g)
        except KeyError:
            pass

        self["groups"] = groups

        return requests

    def setBranch(self, *args, **kwargs):
        """Adds the branch to which the client belongs."""

        try:
            self.mambubranchclass
        except AttributeError:
            from .mambubranch import MambuBranch

            self.mambubranchclass = MambuBranch

        branch = self.mambubranchclass(entid=self["assignedBranchKey"], *args, **kwargs)
        self["assignedBranchName"] = branch["name"]
        self["assignedBranch"] = branch

        return 1

    def create(self, data, *args, **kwargs):
        """Create a client in Mambu

        Parameters
        -data       dictionary with data to send

        """
        return super(MambuClient, self).create(data)

    def update(self, data, *args, **kwargs):
        """Updates a client in Mambu

        Uses PATCH and POST methods, for update "customInformation" and
        "client fields" PATCH method is used, for "customInformation" also
        Patch is used but url needs to be changed, in order to update
        "addresses" or "idDocuments" POST method is needed.

        https://support.mambu.com/docs/clients-api#post-clients
        https://support.mambu.com/docs/clients-api#patch-client
        https://support.mambu.com/docs/clients-api#patch-client-custom-field-values

        Parameters
        -data       dictionary with data to update
        """
        cont_requests = 0
        data2update = {}
        # UPDATE client fields
        if data.get("client"):
            data2update["client"] = data.get("client", {})
            cont_requests += self.update_patch(data2update, *args, **kwargs)

        # UPDATE customFields
        if data.get("customInformation"):
            data2update = {"client": {}}
            data2update["customInformation"] = data.get("customInformation")
            self._MambuStruct__urlfunc = getclientcustominformationurl
            cont_requests += self.update_patch(data2update, *args, **kwargs)
            self._MambuStruct__urlfunc = getclienturl

        # UPDATE addresses or idDocuments
        if data.get("addresses") or data.get("idDocuments"):
            data2update = {}
            # UPDATE addresses
            if data.get("addresses"):
                data2update["addresses"] = data.get("addresses")
            # UPDATE idDocuments
            if data.get("idDocuments"):
                data2update["idDocuments"] = data.get("idDocuments")
            cont_requests += self.update_post(data2update, *args, **kwargs)

        cont_requests += super(MambuClient, self).update(data, *args, **kwargs)

        return cont_requests

    def update_patch(self, data, *args, **kwargs):
        """Updates a client Mambu using method PATCH

        Args:
            data (dictionary): dictionary with data to update

        https://support.mambu.com/docs/clients-api#patch-client
        """
        return super(MambuClient, self).update_patch(data, *args, **kwargs)

    def update_post(self, data, *args, **kwargs):
        """Updates a client in Mambu using method POST

        Args:
            data (dictionary): dictionary with data to update

        https://support.mambu.com/docs/clients-api#post-clients
        """
        return super(MambuClient, self).update_post(data, *args, **kwargs)


class MambuClients(MambuStruct):
    """A list of Clients from Mambu.

    With the default urlfunc, entid argument must be empty at
    instantiation time to retrieve all the clients according to any
    other filter you send to the urlfunc.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """By default, entid argument is empty. That makes perfect
        sense: you want several clients, not just one
        """
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """The trick for iterable Mambu Objects comes here:

                You iterate over each element of the responded List from Mambu,
                and create a Mambu Client object for each one, initializing them
                one at a time, and changing the attrs attribute (which just
                holds a list of plain dictionaries) with a MambuClient just
                created.

        .. todo:: pass a valid (perhaps default) urlfunc, and its
                  corresponding id to entid to each MambuClient, telling
                  MambuStruct not to connect() by default. It's desirable to
                  connect at any other further moment to refresh some element in
                  the list.
        """
        for n, c in enumerate(self.attrs):
            # ok ok, I'm modifying elements of a list while iterating it. BAD PRACTICE!
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            try:
                self.mambuclientclass
            except AttributeError:
                self.mambuclientclass = MambuClient

            client = self.mambuclientclass(urlfunc=None, entid=None, *args, **kwargs)
            client.init(c, *args, **kwargs)
            client._MambuStruct__urlfunc = getclienturl
            self.attrs[n] = client



# coding: utf-8
"""Mambu Groups objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

MambuGroup holds a group.

MambuGroups holds a list of groups.

Uses mambugeturl.getgroupurl as default urlfunc
"""


from ..mambugeturl import getgroupcustominformationurl, getgroupurl
from .mambustruct import MambuStruct
from .mamburestutils import MambuStructIterator

# Other options include getgrouploansurl and getgroupcustominformationurl
mod_urlfunc = getgroupurl


class MambuGroup(MambuStruct):
    """A Group from Mambu.

    With the default urlfunc, entid argument must be the ID of the
    group you wish to retrieve.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """Tasks done here:

        Just initializes the MambuStruct.
        """
        MambuStruct.__init__(
            self, urlfunc, entid, custom_field_name="customInformation", *args, **kwargs
        )

    def preprocess(self):
        """Preprocessing.

                Flattens the object. Important data comes on the 'theGroup'
                dictionary inside of the response. Instead, every element of the
                'theGroup' dictionary is taken out to the main attrs dictionary.

                Notes on the group get some html tags removed.

        .. todo:: use mambuutil.strip_tags() method

                A 'name' field is added, equivalent to the 'groupName' field.
                This is useful on loan accounts. They get a 'holder' field added
                somewhere, which may be an individual client, or a group. So to
                get the holder name, you just access the 'holder'['name']. No
                matter if you have a client loan or a group loan, you get the
                name of the holder.
        """
        super(MambuGroup, self).preprocess()

        try:
            for k, v in self["theGroup"].items():
                self[k] = v
            del self.attrs["theGroup"]
        except Exception:
            pass

        try:
            self["notes"] = self["notes"].replace("<div>", "").replace("</div>", "")
        except Exception:
            self["notes"] = ""

        try:
            self["name"] = self["groupName"]
        except KeyError:
            pass

        self["address"] = {}
        try:
            for name, item in self["addresses"][0].items():
                try:
                    self["addresses"][0][name] = item.strip()
                    self["address"][name] = item.strip()
                except AttributeError:
                    pass
        except (KeyError, IndexError):
            pass

    def setClients(self, *args, **kwargs):
        """Adds the clients for this group to a 'clients' field.

        The 'groupMembers' field of the group holds the encodedKeys of
        the member clients of the group. Since Mambu REST API accepts
        both ids or encodedKeys to retrieve entities, we use that here.

        You may wish to get the full details of each client by passing a
        fullDetails=True argument here.

        Returns the number of requests done to Mambu.
        """

        requests = 0
        if "fullDetails" in kwargs:
            fullDetails = kwargs["fullDetails"]
            kwargs.pop("fullDetails")
        else:
            fullDetails = True

        clients = []

        for m in self["groupMembers"]:
            try:
                self.mambuclientclass = kwargs.pop("mambuclientclass")
            except KeyError:
                try:
                    self.mambuclientclass
                except AttributeError:
                    from .mambuclient import MambuClient

                    self.mambuclientclass = MambuClient

            client = self.mambuclientclass(
                entid=m["clientKey"], fullDetails=fullDetails, *args, **kwargs
            )
            requests += 1
            clients.append(client)

        self["clients"] = clients

        return requests

    def setBranch(self, *args, **kwargs):
        """Adds the branch to which this groups belongs"""
        try:
            self.mambubranchclass
        except AttributeError:
            from .mambubranch import MambuBranch

            self.mambubranchclass = MambuBranch

        branch = self.mambubranchclass(entid=self["assignedBranchKey"], *args, **kwargs)
        self["branch"] = branch

        return 1

    def setCentre(self, *args, **kwargs):
        """Adds the centre to which this groups belongs"""
        try:
            self.mambucentreclass
        except AttributeError:
            from .mambucentre import MambuCentre

            self.mambucentreclass = MambuCentre

        centre = self.mambucentreclass(entid=self["assignedCentreKey"], *args, **kwargs)
        self["centre"] = centre

        return 1

    def setActivities(self, *args, **kwargs):
        """Adds the activities for this group to a 'activities' field.

        Activities are MambuActivity objects.

        Activities get sorted by activity timestamp.

        Returns the number of requests done to Mambu.
        """

        def activity_date(activity):
            """Util function used for sorting activities according to timestamp"""
            try:
                return activity["activity"]["timestamp"]
            except KeyError:
                return None

        try:
            self.mambuactivitiesclass
        except AttributeError:
            from .mambuactivity import MambuActivities

            self.mambuactivitiesclass = MambuActivities

        activities = self.mambuactivitiesclass(
            groupId=self["encodedKey"], *args, **kwargs
        )
        activities.attrs = sorted(activities.attrs, key=activity_date)
        self["activities"] = activities

        return 1

    def create(self, data, *args, **kwargs):
        """Creates a group in Mambu

        Parameters
        -data       dictionary with data to send
        """
        super(MambuGroup, self).create(data)
        if self.get("customInformation"):
            self["group"][self.custom_field_name] = self["customInformation"]
        if self.get("addresses"):
            self["group"]["addresses"] = self["addresses"]
        if self.get("groupMembers"):
            self["group"]["groupMembers"] = self["groupMembers"]
        self.init(attrs=self["group"])
        return 1

    def update(self, data, *args, **kwargs):
        """Updates a group in Mambu

        Uses PATCH and POST methods, for update "groupMembers", "groupRoles"
        and "group fields" PATCH method is used, for "customInformation" also
        Patch is used but url needs to be changed, in order to update
        "addresses" POST method is needed.

        https://support.mambu.com/docs/groups-api#post-group
        https://support.mambu.com/docs/groups-api#patch-group-information
        https://support.mambu.com/docs/groups-api#patch-group-custom-field-values

        Parameters
        -data       dictionary with data to update
        """
        cont_requests = 0
        data2update = {}
        # SET groupMembers
        if data.get("groupMembers"):
            data2update["groupMembers"] = data.get("groupMembers")
        # SET groupRoles
        if data.get("groupRoles"):
            data2update["groupRoles"] = data.get("groupRoles")

        # if there is data to update or data group to update, updates
        if data2update or data.get("group"):
            # SET group fields
            data2update["group"] = data.get("group", {})
            # UPDATE group fields
            cont_requests += self.update_patch(data2update, *args, **kwargs)

        # UPDATE customFields
        if data.get("customInformation"):
            data2update = {"group": {}}
            data2update["customInformation"] = data.get("customInformation")
            self._MambuStruct__urlfunc = getgroupcustominformationurl
            cont_requests += self.update_patch(data2update, *args, **kwargs)
            self._MambuStruct__urlfunc = getgroupurl

        # UPDATE addresses
        if data.get("addresses"):
            data2update = self._notUpdateData()
            data2update["addresses"] = data.get("addresses")
            cont_requests += self.update_post(data2update, *args, **kwargs)

        cont_requests += super(MambuGroup, self).update(data, *args, **kwargs)

        return cont_requests

    def update_patch(self, data, *args, **kwargs):
        """Updates a group Mambu using method PATCH

        Args:
            data (dictionary): dictionary with data to update

        https://support.mambu.com/docs/groups-api#patch-group-information
        """
        return super(MambuGroup, self).update_patch(data, *args, **kwargs)

    def update_post(self, data, *args, **kwargs):
        """Updates a group in Mambu using method POST

        Args:
            data (dictionary): dictionary with data to update

        https://support.mambu.com/docs/groups-api#post-group
        """
        return super(MambuGroup, self).update_post(data, *args, **kwargs)

    def _notUpdateData(self):
        """Data that would not be updated/overwritten in case of POST

        Dictionary with group data copied
        """
        notUpdateFields = [
            "id",
            "groupName",
            "mobilePhone1",
            "homePhone",
            "emailAddress",
            "preferredLanguage",
            "notes",
            "assignedUserKey",
            "assignedCentreKey",
            "assignedBranchKey",
        ]
        data = {"group": {}}
        data["group"] = {f: self.get(f) for f in notUpdateFields if self.get(f)}

        return data

    def addMembers(self, newMembers=[], *args, **kwargs):
        """Adds a list of members to the group

        Args:
            newMembers (list): list with encoded keys of the members to add to group
        """
        data2update = {}
        data2update["group"] = {}
        # get actual members
        groupMembers = [
            {"clientKey": i["clientKey"]} for i in self.get("groupMembers", [])
        ]
        data2update["groupMembers"] = groupMembers
        # add new members if not already exists in group
        newMembers = [
            {"clientKey": i}
            for i in newMembers
            if i not in [c["clientKey"] for c in groupMembers]
        ]
        data2update["groupMembers"].extend(newMembers)

        cont_requests = super(MambuGroup, self).update_patch(data2update)
        self.connect()
        cont_requests += 1
        return cont_requests


class MambuGroups(MambuStruct):
    """A list of Groups from Mambu.

    With the default urlfunc, entid argument must be empty at
    instantiation time to retrieve all the groups according to any
    other filter you send to the urlfunc.
    """

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """By default, entid argument is empty. That makes perfect
        sense: you want several groups, not just one
        """
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """The trick for iterable Mambu Objects comes here:

                You iterate over each element of the responded List from Mambu,
                and create a Mambu Group object for each one, initializing them
                one at a time, and changing the attrs attribute (which just
                holds a list of plain dictionaries) with a MambuGroup just
                created.

        .. todo:: pass a valid (perhaps default) urlfunc, and its
                  corresponding id to entid to each MambuGroup, telling
                  MambuStruct not to connect() by default. It's desirable to
                  connect at any other further moment to refresh some element in
                  the list.
        """
        for n, c in enumerate(self.attrs):
            # ok ok, I'm modifying elements of a list while iterating it. BAD PRACTICE!
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            try:
                self.mambugroupclass
            except AttributeError:
                self.mambugroupclass = MambuGroup

            group = self.mambugroupclass(urlfunc=None, entid=None, *args, **kwargs)
            group.init(c, *args, **kwargs)
            group._MambuStruct__urlfunc = getgroupurl
            self.attrs[n] = group



# coding: utf-8
"""Mambu Roles objects.

.. autosummary::
   :nosignatures:
   :toctree: _autosummary

MambuRole holds a user role.

MambuRoles holds a list of user roles.

Uses mambugeturl.getrolesurl as default urlfunc
"""


from ..mambugeturl import getrolesurl
from .mambustruct import MambuStruct
from .mamburestutils import MambuStructIterator

mod_urlfunc = getrolesurl


class MambuRole(MambuStruct):
    """A Role from Mambu."""

    def __init__(self, urlfunc=mod_urlfunc, entid="", *args, **kwargs):
        """Tasks done here:

        Just initializes the MambuStruct.
        """
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __repr__(self):
        """Instead of the default id given by the parent class, shows
        the rolename
        """
        return self.__class__.__name__ + " - rolename: '%s'" % self["name"]


class MambuRoles(MambuStruct):
    """A list of Roles from Mambu.

    With the default urlfunc, entid argument must be empty at
    instantiation time to retrieve all the roles according to any other
    filter you send to the urlfunc.

    itemclass argument allows you to pass some other class as the
    elements for the list. Why is this useful? You may wish to override
    several behaviours by creating your own MambuRole son class. Pass
    that to the itemclass argument here and voila, you get a list of
    YourMambuRole class using MambuRoles instead of plain old MambuRole
    elements.

    If you wish to specialize other Mambu objects on MambuPy you may
    do that. Mind that if you desire that the iterable version of it
    to have elements of your specialized class, you need to change
    the logic of the constructor and the convert_dict_to_attrs method in
    the iterable class to use some sort of itemclass there too.
    Don't forget to submit the change on a pull request when done
    ;-)
    """

    def __init__(
        self, urlfunc=mod_urlfunc, entid="", itemclass=MambuRole, *args, **kwargs
    ):
        """entid argument is empty. That makes perfect
        sense: you always get serveral roles from Mambu REST API
        """
        self.itemclass = itemclass
        MambuStruct.__init__(self, urlfunc, entid, *args, **kwargs)

    def __iter__(self):
        return MambuStructIterator(self.attrs)

    def convert_dict_to_attrs(self, *args, **kwargs):
        """The trick for iterable Mambu Objects comes here:

        You iterate over each element of the responded List from Mambu,
        and create a Mambu Role (or your own itemclass) object for each
        one, initializing them one at a time, and changing the attrs
        attribute (which just holds a list of plain dictionaries) with a
        MambuUser (or your own itemclass) just created.
        """
        for n, u in enumerate(self.attrs):
            try:
                params = self.params
            except AttributeError:
                params = {}
            kwargs.update(params)
            try:
                self.mamburoleclass
            except AttributeError:
                self.mamburoleclass = MambuRole

            role = self.mamburoleclass(urlfunc=None, entid=None, *args, **kwargs)
            role.init(u, *args, **kwargs)
            role._MambuStruct__urlfunc = getrolesurl
            self.attrs[n] = role



from ..mambuutil import MambuError


class RequestsCounter(object):
    """Singleton that counts requests.

    If you want to control the number of requests you do to Mambu, you
    may find this counter useful. Since it is a Singleton, every Mambu
    object shares it and increases the amount of requests counted here,
    so you may read it on every Mambu object you have per Python
    session you hold.
    """
    __instance = None

    def __new__(cls, *args, **kwargs):
        if not cls.__instance:
            cls.__instance = super(RequestsCounter, cls).__new__(cls, *args, **kwargs)
            cls.requests = []
            cls.cnt = 0
        return cls.__instance

    def add(cls, temp):
        cls.requests.append(temp)
        cls.cnt += 1

    def reset(cls):
        cls.cnt = 0
        cls.requests = []


class MambuStructIterator:
    """Enables iteration for some Mambu objects that may want to have iterators.

    Loans, Transactions, Repayments, Clients, Groups, Users, Branches
    are some of the Mambu entitites that may be iterable. Here at
    MambuPy, all of them have an iterator class for requesting several
    entitites to Mambu, instead of just one of them. Please refer to
    each MambuObject pydoc for more info.
    """

    def __init__(self, wrapped):
        self.wrapped = wrapped
        self.offset = 0

    def __next__(self):
        if self.offset >= len(self.wrapped):
            raise StopIteration
        else:
            item = self.wrapped[self.offset]
            self.offset += 1
            return item

    def next(self):
        return self.__next__()

    def __iter__(self):
        return self


def set_custom_field(mambuentity, customfield="", *args, **kwargs):
    """Modifies the customField field for the given object with
    something related to the value of the given field.

    If the dataType == "USER_LINK" then instead of using the value
    of the CF, it will be a MambuUser object.

    Same if dataType == "CLIENT_LINK", but with a MambuClient.

    Default case: just uses the same value the CF already had.

    Returns the number of requests done to Mambu.
    """
    from . import mambuclient, mambuuser

    try:
        customFieldValue = mambuentity[customfield]
        # find the dataType customfield by name or id
        datatype = [
            l["customField"]["dataType"]
            for l in mambuentity[mambuentity.custom_field_name]
            if (l["name"] == customfield or l["id"] == customfield)
        ][0]
    except IndexError:
        # if no customfield found with the given name, assume it is a grouped
        # custom field, name must have an index suffix that must be removed
        try:
            # find the dataType customfield by name or id
            datatype = [
                l["customField"]["dataType"]
                for l in mambuentity[mambuentity.custom_field_name]
                if (
                    l["name"] == customfield.split("_")[0] or
                    l["id"] == customfield.split("_")[0]
                )
            ][0]
        except IndexError:
            err = MambuError(
                "Object %s has no custom field '%s'" % (mambuentity["id"], customfield)
            )
            raise err
    except AttributeError:
        err = MambuError("Object does not have a custom field to set")
        raise err

    if datatype == "USER_LINK":
        mambuentity[customfield] = mambuuser.MambuUser(
            entid=customFieldValue, *args, **kwargs
        )
    elif datatype == "CLIENT_LINK":
        mambuentity[customfield] = mambuclient.MambuClient(
            entid=customFieldValue, *args, **kwargs
        )
    else:
        mambuentity[customfield] = customFieldValue
        return 0
    return 1
